unit ALU;
interface

uses Tools, RAM;

Type TFlags=(fCarry,fAux,fF0,fRS1,fRS0,fOverflow,fF1,fParity);

Type TFlagSet=set of TFlags;

Type TSeriellOutputEvent=Procedure (Sender : TObject; Value : Byte) of object;

Type TAnalogReadEvent=Procedure (Sender : TObject; Nr : Byte; Var Value : Integer) of object;

Type TRegister=class
  private
    FData : TSpeicher;
    FSFR, FSFR2 : TSFRArea;
    SeriellIRQ, CapComIRQ1, CapComIRQ2 : Boolean;
    Procedure SetR(I : Integer; B : Byte);
    Function GetR(I : Integer) : Byte;
    Procedure SetFlags(F : TFlagSet);
    Function GetFlags : TFlagSet;
    Procedure SetByte(Index : Integer; B : Byte);
    Function GetByte(Index : Integer) : Byte;
    Procedure SetWord(Index : Integer; W : Word);
    Function GetWord(Index : Integer) : Word;
    Function GetSFRBit(Index : Integer) : Boolean;
    Procedure SetSFRBit(Index : Integer; B : Boolean);
    function GetWaitingIRQ(I : Integer) : Boolean;
    procedure SetWaitingIRQ(I : Integer; Value : Boolean);
  public
    PC : Word;
    LowLovelBefore : Boolean; {wenn High-Priority einen Low-Priority unterbrochen hat}
    IRQLevel : Byte; {0=alle zulassen, 1=Low-Priority in Arbeit, 2=High-Priority in Arbeit}
    Constructor Create(const Data : TSpeicher; SFR, SFR2 : TSFRArea);
    Procedure Init(Addr : Integer = 0);
    Procedure SetACC(I : Integer);
    Procedure AddToACC(B : Byte; WithCarry : Boolean = False);
    Procedure SubFromACC(B : Byte); {Immer mit Carry}
    Procedure BCDCorrectAkku;
    Procedure CalcParity;
    Procedure RequestIRQ(Nr : Byte);
    Function NextWaitingIRQ : Integer;
    property Flags : TFlagSet read GetFlags write SetFlags;
    property R[I : Integer] : Byte read GetR write SetR;
    property ACC : Byte index 1 read GetByte write SetByte;
    property B : Byte index 2 read GetByte write SetByte;
    property PSW : Byte index 3 read GetByte write SetByte;
    property SP : Byte index 4 read GetByte write SetByte;
    property DPTR : Word index 1 read GetWord write SetWord;
    property XMAP : Boolean index 1 read GetSFRBit write SetSFRBit;
    property RMAP : Boolean index 2 read GetSFRBit write SetSFRBit;
    property WaitingIRQ[I : Integer] : Boolean read GetWaitingIRQ write SetWaitingIRQ;
end;

Type TProzessor=class
  private
    FRegister : TRegister;
    FExt, FData, FCode, FXRAM : TSpeicher;
    FSFRArea, FSFRArea2 : TSFRArea;
    LastT0, LastT1 : Boolean;
    LastINT : Array [0..1] of Boolean;
    SeriellCounter : Integer;
    FSeriellStep : Integer;
    TaktCounter : Integer;
    LastCapComRunning, RunningUp : Boolean;
    FOnSeriellOutput : TSeriellOutputEvent;
    FOnReadAnalog : TAnalogReadEvent;
    Function GetDataDirect(B : Byte; ReadModifyWrite : Boolean) : Byte;
    Procedure SetDataDirect(Addr, B : Byte; ReadModifyWrite : Boolean);
    Function GetDataIndirect(B : Byte) : Byte;
    Procedure SetDataIndirect(Addr, B : Byte);
    Function GetDataExtern(Addr : Word) : Byte;
    Procedure SetDataExtern(Addr : Word; B : Byte);
    Function GetDataExternIndirect(B : Byte) : Byte;
    Procedure SetDataExternIndirect(Addr, B : Byte);
    Function GetBit(Nr : Byte; ReadModifyWrite : Boolean) : Boolean;
    Procedure SetBit(Nr : Byte; B : Boolean; ReadModifyWrite : Boolean);
    Function GetCarry : Boolean;
    Procedure SetCarry(B : Boolean);
    Procedure DoStep;
    Procedure StartIRQ(Nr : Integer);
    Procedure TimerWork;
    Procedure IncTimer(Nr, Mode  : Byte; Mode3 : Boolean);
    Procedure SeriellWork;
    Procedure SendSBUF(Sender : TObject);
    Procedure ExtIntWork;
    Procedure CapComWork;
    Procedure SetCapComPort(Nr : Byte; Value : Boolean; CountingDown : Boolean = False);
    Procedure ADWork;
  public
    SingleMemory1,SingleMemory2 : Boolean;
    T0Click, T1Click : Boolean;
    SeriellInBuffer : String;
    Constructor Create(const ACode, AExt, AData : TSpeicher; ASFRArea, ASFRArea2 : TSFRArea; AXRAM : TSpeicher);
    Destructor Destroy; override;
    function Step : Boolean;
    property Reg : TRegister read FRegister write FRegister;
    property Data : TSpeicher read FData;
    property Code : TSpeicher read FCode;
    property SFR : TSFRArea read FSFRArea;
    property SeriellStep : Integer read FSeriellStep write FSeriellStep;
  published
    property OnSeriellOutput : TSeriellOutputEvent read FOnSeriellOutput write FOnSeriellOutput;
    property OnReadAnalog : TAnalogReadEvent read FOnReadAnalog write FOnReadAnalog;
end;

implementation

uses SysUtils, PrgConst, Classes;

Const RMAPTab : Array[0..23] of Byte = (
    $88,
    $90,
    $B0,$BF,
    $D2,$D3,$D4,$D5,$D6,$D7,$DE,$DF,
    $E2,$E3,$E4,$E5,$E6,$E7,
    $F2,$F3,$F4,$F5,$F6,$F7
  );

Const Long2Ops : Array[0..91] of Byte = (
    $01,
    $02,
    $10,
    $11,
    $12,
    $20,
    $21,
    $22,
    $30,
    $31,
    $32,
    $40,
    $41,
    $43,
    $50,
    $51,
    $53,
    $60,
    $61,
    $63,
    $70,
    $71,
    $72,
    $73,
    $75,
    $80,
    $81,
    $82,
    $83,
    $85,
    $86,$87,
    $88,$89,$8A,$8B,$8C,$8D,$8E,$8F,
    $90,
    $91,
    $92,
    $93,
    $A0,
    $A1,
    $A3,
    $A6,$A7,
    $A8,$A9,$AA,$AB,$AC,$AD,$AE,$AF,
    $B0,
    $B1,
    $B4,
    $B5,
    $B6,$B7,
    $B8,$B9,$BA,$BB,$BC,$BD,$BE,$BF,
    $C0,
    $C1,
    $D0,
    $D1,
    $D5,
    $D8,$D9,$DA,$DB,$DC,$DD,$DE,$DF,
    $E0,
    $E1,
    $E2,$E3,
    $F0,
    $F1,
    $F2,$F3
  );

{ TProzessor }

constructor TProzessor.Create(const ACode, AExt, AData : TSpeicher; ASFRArea, ASFRArea2 : TSFRArea; AXRAM : TSpeicher);
Var x : Integer;
begin
  inherited Create;
  FCode:=ACode; FExt:=AExt; FData:=AData; FSFRArea:=ASFRArea; FSFRArea2:=ASFRArea2; FXRAM:=AXRAM;
  FRegister:=TRegister.Create(FData,FSFRArea,FSFRArea2);
  LastT0:=False; LastT1:=False; FSeriellStep:=1000;
  For X:=0 to 1 do LastINT[X]:=True;
  FSFRArea.OnSBUFWrite:=SendSBUF;
  RunningUp:=True;
end;

destructor TProzessor.Destroy;
begin
  FRegister.Free;
  inherited Destroy;
end;

function TProzessor.GetDataDirect(B: Byte; ReadModifyWrite : Boolean): Byte;
Var X : Integer;
begin
  If B<128 then result:=FData[B] else begin
    If FRegister.RMAP then For X:=0 to High(RMAPTab) do If B=RMAPTab[X] then begin
      If ReadModifyWrite then result:=FSFRArea2.Read2(B) else result:=FSFRArea2[B];
      exit;
    end;
    If ReadModifyWrite then result:=FSFRArea.Read2(B) else result:=FSFRArea[B];
  end;
end;

procedure TProzessor.SetDataDirect(Addr, B : Byte; ReadModifyWrite : Boolean);
Var X : Integer;
begin
  If Addr<128 then FData[Addr]:=B else begin
    If FRegister.RMAP then For X:=0 to High(RMAPTab) do If Addr=RMAPTab[X] then begin
      FSFRArea2[Addr]:=B; exit;
    end;
    FSFRArea[Addr]:=B;
  end;
end;

function TProzessor.GetDataIndirect(B: Byte): Byte;
begin
  result:=FData[FRegister.R[B]];
end;

procedure TProzessor.SetDataIndirect(Addr, B : Byte);
begin
  FData[FRegister.R[Addr]]:=B;
end;

Procedure TProzessor.DoStep;
Var I,len : Integer;
    B,C : Byte;
    W,W2 : word;
begin
  len:=1;
  try
    Case FCode[FRegister.PC] of
            {nop}
      $00 : ;
            {ajmp 0rel}
      $01 : begin FRegister.PC:=((FRegister.PC+2) and $E000)+0*256+FCode[FRegister.PC+1]; len:=0; end;
            {ljmp <Adr16>}
      $02 : begin FRegister.PC:=256*FCode[FRegister.PC+1]+FCode[FRegister.PC+2]; len:=0; end;
            {rr a}
      $03 : begin B:=FRegister.ACC; asm ror B,1; end; FRegister.ACC:=B; end;
            {inc a}
      $04 : begin B:=FRegister.ACC; inc(B); FRegister.ACC:=B; end;
            {inc dadr}
      $05 : begin B:=GetDataDirect(FCode[FRegister.PC+1],True); inc(B); SetDataDirect(FCode[FRegister.PC+1],B,True); len:=2; end;
            {inc @ri}
      $06 : begin B:=GetDataIndirect(0); inc(B); SetDataIndirect(0,B); end;
      $07 : begin B:=GetDataIndirect(1); inc(B); SetDataIndirect(1,B); end;
            {inc rx}
      $08 : begin B:=FRegister.R[0]; inc(b); FRegister.R[0]:=B; end;
      $09 : begin B:=FRegister.R[1]; inc(b); FRegister.R[1]:=B; end;
      $0A : begin B:=FRegister.R[2]; inc(b); FRegister.R[2]:=B; end;
      $0B : begin B:=FRegister.R[3]; inc(b); FRegister.R[3]:=B; end;
      $0C : begin B:=FRegister.R[4]; inc(b); FRegister.R[4]:=B; end;
      $0D : begin B:=FRegister.R[5]; inc(b); FRegister.R[5]:=B; end;
      $0E : begin B:=FRegister.R[6]; inc(b); FRegister.R[6]:=B; end;
      $0F : begin B:=FRegister.R[7]; inc(b); FRegister.R[7]:=B; end;
            {jbc badr,rel}
      $10 : If GetBit(FCode[FRegister.PC+1],False) then begin
              SetBit(FCode[FRegister.PC+1],False,False); len:=3+ShortInt(FCode[FRegister.PC+2])
            end else len:=3;
            {acall 0rel}
      $11 : begin
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) mod 256;
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) div 256;
              FRegister.PC:=((FRegister.PC+2) and $E000)+0*256+FCode[FRegister.PC+1]; len:=0;
            end;
            {lcall #const16}
      $12 : begin
              W2:=FCode[FRegister.PC+1]*256+FCode[FRegister.PC+2]; inc(FRegister.PC,3);
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=FRegister.PC mod 256;
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=FRegister.PC div 256;
              FRegister.PC:=W2; len:=0;
            end;
            {rrc a}
      $13 : begin
              I:=FRegister.ACC; If fCarry in FRegister.Flags then B:=128 else B:=0;
              SetCarry(I mod 2 = 1); I:=I div 2; FRegister.ACC:=I+B;
            end;
            {dec a}
      $14 : begin B:=FRegister.ACC; dec(B); FRegister.ACC:=B; end;
            {dec dadr}
      $15 : begin B:=GetDataDirect(FCode[FRegister.PC+1],True); dec(B); SetDataDirect(FCode[FRegister.PC+1],B,True); len:=2; end;
            {dec @ri}
      $16 : begin B:=GetDataIndirect(0); dec(B); SetDataIndirect(0,B); end;
      $17 : begin B:=GetDataIndirect(1); dec(B); SetDataIndirect(1,B); end;
            {dec rx}
      $18 : begin B:=FRegister.R[0]; dec(B); FRegister.R[0]:=B; end;
      $19 : begin B:=FRegister.R[1]; dec(B); FRegister.R[1]:=B; end;
      $1A : begin B:=FRegister.R[2]; dec(B); FRegister.R[2]:=B; end;
      $1B : begin B:=FRegister.R[3]; dec(B); FRegister.R[3]:=B; end;
      $1C : begin B:=FRegister.R[4]; dec(B); FRegister.R[4]:=B; end;
      $1D : begin B:=FRegister.R[5]; dec(B); FRegister.R[5]:=B; end;
      $1E : begin B:=FRegister.R[6]; dec(B); FRegister.R[6]:=B; end;
      $1F : begin B:=FRegister.R[7]; dec(B); FRegister.R[7]:=B; end;
            {jb badr,rel}
      $20 : If GetBit(FCode[FRegister.PC+1],False) then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            {ajmp 1rel}
      $21 : begin FRegister.PC:=((FRegister.PC+2) and $E000)+1*256+FCode[FRegister.PC+1]; len:=0; end;
            {ret}
      $22 : begin
              B:=FRegister.SP; FRegister.PC:=FData[B]*256+FData[B-1];
              dec(B,2); FRegister.SP:=B; len:=0;
            end;
            {rl a}
      $23 : begin B:=FRegister.ACC; asm rol B,1; end; FRegister.ACC:=B; end;
            {add a,#const8}
      $24 : begin FRegister.AddToACC(FCode[FRegister.PC+1]); len:=2; end;
            {add a,dadr}
      $25 : begin FRegister.AddToACC(GetDataDirect(FCode[FRegister.PC+1],False)); len:=2; end;
            {add a,@ri}
      $26 : FRegister.AddToACC(GetDataIndirect(0));
      $27 : FRegister.AddToACC(GetDataIndirect(1));
            {add a,rx}
      $28 : FRegister.AddToACC(FRegister.R[0]);
      $29 : FRegister.AddToACC(FRegister.R[1]);
      $2A : FRegister.AddToACC(FRegister.R[2]);
      $2B : FRegister.AddToACC(FRegister.R[3]);
      $2C : FRegister.AddToACC(FRegister.R[4]);
      $2D : FRegister.AddToACC(FRegister.R[5]);
      $2E : FRegister.AddToACC(FRegister.R[6]);
      $2F : FRegister.AddToACC(FRegister.R[7]);
            {jnb badr,rel}
      $30 : If not GetBit(FCode[FRegister.PC+1],False) then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            {acall 1rel}
      $31 : begin
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) mod 256;
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) div 256;
              FRegister.PC:=((FRegister.PC+2) and $E000)+1*256+FCode[FRegister.PC+1]; len:=0;
            end;
            {reti}
      $32 : begin
              B:=FRegister.SP; FRegister.PC:=FData[B]*256+FData[B-1];
              dec(B,2); FRegister.SP:=B; len:=0;
              Case FRegister.IRQLevel of
                3 : FRegister.IRQLevel:=1;
                2 : If FRegister.LowLovelBefore then begin
                      FRegister.IRQLevel:=1; FRegister.LowLovelBefore:=False;
                    end else FRegister.IRQLevel:=0;
                1 : FRegister.IRQLevel:=0;
              end;
            end;
            {rlc a}
      $33 : begin
              I:=FRegister.ACC; If fCarry in FRegister.Flags then B:=1 else B:=0;
              SetCarry(I>=128); I:=(I*2) mod 256; FRegister.ACC:=I+B;
            end;
            {addc a,#const8}
      $34 : begin FRegister.AddToACC(FCode[FRegister.PC+1],True); len:=2; end;
            {addc a,dadr}
      $35 : begin FRegister.AddToACC(GetDataDirect(FCode[FRegister.PC+1],False),True); len:=2; end;
            {addc a,@ri}
      $36 : FRegister.AddToACC(GetDataIndirect(0),True);
      $37 : FRegister.AddToACC(GetDataIndirect(1),True);
            {addc a,rx}
      $38 : FRegister.AddToACC(FRegister.R[0],True);
      $39 : FRegister.AddToACC(FRegister.R[1],True);
      $3A : FRegister.AddToACC(FRegister.R[2],True);
      $3B : FRegister.AddToACC(FRegister.R[3],True);
      $3C : FRegister.AddToACC(FRegister.R[4],True);
      $3D : FRegister.AddToACC(FRegister.R[5],True);
      $3E : FRegister.AddToACC(FRegister.R[6],True);
      $3F : FRegister.AddToACC(FRegister.R[7],True);
            {jc rel}
      $40 : If fCarry in FRegister.Flags then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            {ajmp 2rel}
      $41 : begin FRegister.PC:=((FRegister.PC+2) and $E000)+2*256+FCode[FRegister.PC+1]; len:=0; end;
            {orl dadr,a}
      $42 : begin SetDataDirect(FCode[FRegister.PC+1],GetDataDirect(FCode[FRegister.PC+1],True) or FRegister.ACC,True); len:=2; end;
            {orl dadr,#const8}
      $43 : begin SetDataDirect(FCode[FRegister.PC+1],GetDataDirect(FCode[FRegister.PC+1],True) or FCode[FRegister.PC+2],True); len:=3; end;
            {orl a,#const8}
      $44 : begin FRegister.ACC:=FRegister.ACC or FCode[FRegister.PC+1]; len:=2; end;
            {orl a,dadr}
      $45 : begin FRegister.ACC:=FRegister.ACC or GetDataDirect(FCode[FRegister.PC+1],True); len:=2; end;
            {orl a,@ri}
      $46 : FRegister.ACC:=FRegister.ACC or GetDataIndirect(0);
      $47 : FRegister.ACC:=FRegister.ACC or GetDataIndirect(1);
            {orl a,rx}
      $48 : FRegister.ACC:=FRegister.ACC or FRegister.R[0];
      $49 : FRegister.ACC:=FRegister.ACC or FRegister.R[1];
      $4A : FRegister.ACC:=FRegister.ACC or FRegister.R[2];
      $4B : FRegister.ACC:=FRegister.ACC or FRegister.R[3];
      $4C : FRegister.ACC:=FRegister.ACC or FRegister.R[4];
      $4D : FRegister.ACC:=FRegister.ACC or FRegister.R[5];
      $4E : FRegister.ACC:=FRegister.ACC or FRegister.R[6];
      $4F : FRegister.ACC:=FRegister.ACC or FRegister.R[7];
            {jnc rel}
      $50 : If not (fCarry in FRegister.Flags) then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            {acall 2rel}
      $51 : begin
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) mod 256;
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) div 256;
              FRegister.PC:=((FRegister.PC+2) and $E000)+2*256+FCode[FRegister.PC+1]; len:=0;
            end;
            {anl dadr,a}
      $52 : begin SetDataDirect(FCode[FRegister.PC+1],GetDataDirect(FCode[FRegister.PC+1],True) and FRegister.ACC,True); len:=2; end;
            {anl dadr,#const8}
      $53 : begin SetDataDirect(FCode[FRegister.PC+1],GetDataDirect(FCode[FRegister.PC+1],True) and FCode[FRegister.PC+2],True); len:=3; end;
            {anl a,#const8}
      $54 : begin FRegister.ACC:=FRegister.ACC and FCode[FRegister.PC+1]; len:=2; end;
            {anl a,dadr}
      $55 : begin FRegister.ACC:=FRegister.ACC and GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
            {anl a,@ri}
      $56 : FRegister.ACC:=FRegister.ACC and GetDataIndirect(0);
      $57 : FRegister.ACC:=FRegister.ACC and GetDataIndirect(1);
            {anl a,rx}
      $58 : FRegister.ACC:=FRegister.ACC and FRegister.R[0];
      $59 : FRegister.ACC:=FRegister.ACC and FRegister.R[1];
      $5A : FRegister.ACC:=FRegister.ACC and FRegister.R[2];
      $5B : FRegister.ACC:=FRegister.ACC and FRegister.R[3];
      $5C : FRegister.ACC:=FRegister.ACC and FRegister.R[4];
      $5D : FRegister.ACC:=FRegister.ACC and FRegister.R[5];
      $5E : FRegister.ACC:=FRegister.ACC and FRegister.R[6];
      $5F : FRegister.ACC:=FRegister.ACC and FRegister.R[7];
            {jz rel}
      $60 : If FRegister.ACC=0 then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            {ajmp 3rel}
      $61 : begin FRegister.PC:=((FRegister.PC+2) and $E000)+3*256+FCode[FRegister.PC+1]; len:=0; end;
            {xrl dadr,a}
      $62 : begin SetDataDirect(FCode[FRegister.PC+1],GetDataDirect(FCode[FRegister.PC+1],True) xor FRegister.ACC,True); len:=2; end;
            {xrl dadr,#const8}
      $63 : begin SetDataDirect(FCode[FRegister.PC+1],GetDataDirect(FCode[FRegister.PC+1],True) xor FCode[FRegister.PC+2],True); len:=3; end;
            {xrl a,#const8}
      $64 : begin FRegister.ACC:=FRegister.ACC xor FCode[FRegister.PC+1]; len:=2; end;
            {xrl a,dadr}
      $65 : begin FRegister.ACC:=FRegister.ACC xor GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
            {xrl a,@ri}
      $66 : FRegister.ACC:=FRegister.ACC xor GetDataIndirect(0);
      $67 : FRegister.ACC:=FRegister.ACC xor GetDataIndirect(1);
            {xrl a,rx}
      $68 : FRegister.ACC:=FRegister.ACC xor FRegister.R[0];
      $69 : FRegister.ACC:=FRegister.ACC xor FRegister.R[1];
      $6A : FRegister.ACC:=FRegister.ACC xor FRegister.R[2];
      $6B : FRegister.ACC:=FRegister.ACC xor FRegister.R[3];
      $6C : FRegister.ACC:=FRegister.ACC xor FRegister.R[4];
      $6D : FRegister.ACC:=FRegister.ACC xor FRegister.R[5];
      $6E : FRegister.ACC:=FRegister.ACC xor FRegister.R[6];
      $6F : FRegister.ACC:=FRegister.ACC xor FRegister.R[7];
            {jnz rel}
      $70 : If FRegister.ACC<>0 then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            {acall 3rel}
      $71 : begin
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) mod 256;
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) div 256;
              FRegister.PC:=((FRegister.PC+2) and $E000)+3*256+FCode[FRegister.PC+1]; len:=0;
            end;
            {orl c,badr}
      $72 : begin SetCarry(GetBit(FCode[FRegister.PC+1],False) or GetCarry); len:=2; end;
            {jmp @a+dptr}
      $73 : begin FRegister.PC:=FRegister.ACC+FRegister.DPTR; len:=0; end;
            {mov a,#const8}
      $74 : begin FRegister.ACC:=FCode[FRegister.PC+1]; len:=2; end;
            {mov dadr,#const8}
      $75 : begin SetDataDirect(FCode[FRegister.PC+1],FCode[FRegister.PC+2],False); len:=3; end;
            {mov @ri,#const8}
      $76 : begin SetDataIndirect(0,FCode[FRegister.PC+1]); len:=2; end;
      $77 : begin SetDataIndirect(1,FCode[FRegister.PC+1]); len:=2; end;
            {mov rx,#const8}
      $78 : begin FRegister.R[0]:=FCode[FRegister.PC+1]; len:=2; end;
      $79 : begin FRegister.R[1]:=FCode[FRegister.PC+1]; len:=2; end;
      $7A : begin FRegister.R[2]:=FCode[FRegister.PC+1]; len:=2; end;
      $7B : begin FRegister.R[3]:=FCode[FRegister.PC+1]; len:=2; end;
      $7C : begin FRegister.R[4]:=FCode[FRegister.PC+1]; len:=2; end;
      $7D : begin FRegister.R[5]:=FCode[FRegister.PC+1]; len:=2; end;
      $7E : begin FRegister.R[6]:=FCode[FRegister.PC+1]; len:=2; end;
      $7F : begin FRegister.R[7]:=FCode[FRegister.PC+1]; len:=2; end;
            {anl c,badr}
      $82 : begin SetCarry(GetBit(FCode[FRegister.PC+1],False) and GetCarry); len:=2; end;
            {ajmp 4rel}
      $81 : begin FRegister.PC:=((FRegister.PC+2) and $E000)+4*256+FCode[FRegister.PC+1]; len:=0; end;
            {sjmp rel}
      $80 : len:=2+ShortInt(FCode[FRegister.PC+1]);
            {movc a,@a+pc}
      $83 : FRegister.ACC:=FCode[FRegister.ACC+FRegister.PC+1];
            {div ab}
      $84 : begin
              SetCarry(False);
              If FRegister.B=0 then begin
                FRegister.Flags:=FRegister.Flags+[fOverflow];
              end else begin
                FRegister.Flags:=FRegister.Flags-[fOverflow];
                B:=FRegister.ACC div FRegister.B;
                C:=FRegister.ACC mod FRegister.B;
                FRegister.ACC:=B; FRegister.B:=C;
              end;
            end;
            {mov dadr, dadr}
      $85 : begin SetDataDirect(FCode[FRegister.PC+1],GetDataDirect(FCode[FRegister.PC+2],False),False); len:=3; end;
            {mov dadr,@ri}
      $86 : begin SetDataDirect(FCode[FRegister.PC+1],GetDataIndirect(0),False); len:=2; end;
      $87 : begin SetDataDirect(FCode[FRegister.PC+1],GetDataIndirect(1),False); len:=2; end;
            {mov dadr,rx}
      $88 : begin SetDataDirect(FCode[FRegister.PC+1],FRegister.R[0],False); len:=2; end;
      $89 : begin SetDataDirect(FCode[FRegister.PC+1],FRegister.R[1],False); len:=2; end;
      $8A : begin SetDataDirect(FCode[FRegister.PC+1],FRegister.R[2],False); len:=2; end;
      $8B : begin SetDataDirect(FCode[FRegister.PC+1],FRegister.R[3],False); len:=2; end;
      $8C : begin SetDataDirect(FCode[FRegister.PC+1],FRegister.R[4],False); len:=2; end;
      $8D : begin SetDataDirect(FCode[FRegister.PC+1],FRegister.R[5],False); len:=2; end;
      $8E : begin SetDataDirect(FCode[FRegister.PC+1],FRegister.R[6],False); len:=2; end;
      $8F : begin SetDataDirect(FCode[FRegister.PC+1],FRegister.R[7],False); len:=2; end;
            {mov DPRT,#const16}
      $90 : begin FRegister.DPTR:=FCode[FRegister.PC+1]*256+FCode[FRegister.PC+2]; len:=3; end;
            {acall 4rel}
      $91 : begin
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) mod 256;
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) div 256;
              FRegister.PC:=((FRegister.PC+2) and $E000)+4*256+FCode[FRegister.PC+1]; len:=0;
            end;
            {mov badr,c}
      $92 : begin SetBit(FCode[FRegister.PC+1],GetCarry,False); len:=2; end;
            {movc a,@a+dptr}
      $93 : FRegister.ACC:=FCode[FRegister.ACC+FRegister.DPTR];
            {subb a,#const8}
      $94 : begin FRegister.SubFromACC(FCode[FRegister.PC+1]); len:=2; end;
            {subb a,dadr}
      $95 : begin FRegister.SubFromACC(GetDataDirect(FCode[FRegister.PC+1],False)); len:=2; end;
            {subb a,@ri}
      $96 : FRegister.SubFromACC(GetDataIndirect(0));
      $97 : FRegister.SubFromACC(GetDataIndirect(1));
            {subb a,rx}
      $98 : FRegister.SubFromACC(FRegister.R[0]);
      $99 : FRegister.SubFromACC(FRegister.R[1]);
      $9A : FRegister.SubFromACC(FRegister.R[2]);
      $9B : FRegister.SubFromACC(FRegister.R[3]);
      $9C : FRegister.SubFromACC(FRegister.R[4]);
      $9D : FRegister.SubFromACC(FRegister.R[5]);
      $9E : FRegister.SubFromACC(FRegister.R[6]);
      $9F : FRegister.SubFromACC(FRegister.R[7]);
            {orl c,/badr}
      $A0 : begin SetCarry((not GetBit(FCode[FRegister.PC+1],False)) or GetCarry); len:=2; end;
            {ajmp 5rel}
      $A1 : begin FRegister.PC:=((FRegister.PC+2) and $E000)+5*256+FCode[FRegister.PC+1]; len:=0; end;
            {mov c,badr}
      $A2 : begin SetCarry(GetBit(FCode[FRegister.PC+1],False)); len:=2; end;
            {inc DPTR}
      $A3 : begin W:=FRegister.DPTR; inc(W); FRegister.DPTR:=W; end;
            {mul AB}
      $A4 : begin
              I:=FRegister.ACC*FRegister.B; SetCarry(False);
              If I>255 then FRegister.Flags:=FRegister.Flags+[fOverflow] else FRegister.Flags:=FRegister.Flags-[fOverflow];
              FRegister.ACC:=I mod 256; FRegister.B:=I div 256;
            end;
      {A5 : gibt's nicht}
            {mov @ri,dadr}
      $A6 : begin SetDataIndirect(0,GetDataDirect(FCode[FRegister.PC+1],False)); len:=2; end;
      $A7 : begin SetDataIndirect(1,GetDataDirect(FCode[FRegister.PC+1],False)); len:=2; end;
            {mov rx,dadr}
      $A8 : begin FRegister.R[0]:=GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
      $A9 : begin FRegister.R[1]:=GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
      $AA : begin FRegister.R[2]:=GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
      $AB : begin FRegister.R[3]:=GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
      $AC : begin FRegister.R[4]:=GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
      $AD : begin FRegister.R[5]:=GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
      $AE : begin FRegister.R[6]:=GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
      $AF : begin FRegister.R[7]:=GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
            {anl c,/badr}
      $B0 : begin SetCarry((not GetBit(FCode[FRegister.PC+1],False)) and GetCarry); len:=2; end;
            {acall 5rel}
      $B1 : begin
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) mod 256;
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) div 256;
              FRegister.PC:=((FRegister.PC+2) and $E000)+5*256+FCode[FRegister.PC+1]; len:=0;
            end;
            {cpl badr}
      $B2 : begin SetBit(FCode[FRegister.PC+1],not GetBit(FCode[FRegister.PC+1],True),True); len:=2; end;
            {cpl c}
      $B3 : SetCarry(not GetCarry);
            {cjne a,#const8,rel}
      $B4 : begin
              I:=FRegister.ACC-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
            {cjne a,dadr,rel}
      $B5 : begin
              I:=FRegister.ACC-GetDataDirect(FCode[FRegister.PC+1],False); SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
            {cjne @ri,dadr,rel}
      $B6 : begin
              I:=GetDataIndirect(0)-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
      $B7 : begin
              I:=GetDataIndirect(0)-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
            {cjne rx,dadr,rel}
      $B8 : begin
              I:=FRegister.R[0]-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
      $B9 : begin
              I:=FRegister.R[1]-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
      $BA : begin
              I:=FRegister.R[2]-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
      $BB : begin
              I:=FRegister.R[3]-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
      $BC : begin
              I:=FRegister.R[4]-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
      $BD : begin
              I:=FRegister.R[5]-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
      $BE : begin
              I:=FRegister.R[6]-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
      $BF : begin
              I:=FRegister.R[7]-FCode[FRegister.PC+1]; SetCarry(I<0);
              If I<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
            {push dadr}
      $C0 : begin
              B:=FRegister.SP; inc(B); FRegister.SP:=B;
              FData[B]:=GetDataDirect(FCode[FRegister.PC+1],False); len:=2;
            end;
            {ajmp 6rel}
      $C1 : begin FRegister.PC:=((FRegister.PC+2) and $E000)+6*256+FCode[FRegister.PC+1]; len:=0; end;
            {clr badr}
      $C2 : begin SetBit(FCode[FRegister.PC+1],False,False); len:=2; end;
            {clr c}
      $C3 : SetCarry(False);
            {swap a}
      $C4 : FRegister.ACC:=(FRegister.ACC div 16)+(FRegister.ACC*16) mod 256;
            {xch a,dadr}
      $C5 : begin B:=FRegister.ACC; FRegister.ACC:=GetDataDirect(FCode[FRegister.PC+1],False); SetDataDirect(FCode[FRegister.PC+1],B,False); len:=2; end;
            {xch a,@ri}
      $C6 : begin B:=FRegister.ACC; FRegister.ACC:=GetDataIndirect(0); SetDataIndirect(0,B); end;
      $C7 : begin B:=FRegister.ACC; FRegister.ACC:=GetDataIndirect(1); SetDataIndirect(1,B); end;
            {xch a,rx}
      $C8 : begin B:=FRegister.ACC; FRegister.ACC:=FRegister.R[0]; FRegister.R[0]:=B; end;
      $C9 : begin B:=FRegister.ACC; FRegister.ACC:=FRegister.R[1]; FRegister.R[1]:=B; end;
      $CA : begin B:=FRegister.ACC; FRegister.ACC:=FRegister.R[2]; FRegister.R[2]:=B; end;
      $CB : begin B:=FRegister.ACC; FRegister.ACC:=FRegister.R[3]; FRegister.R[3]:=B; end;
      $CC : begin B:=FRegister.ACC; FRegister.ACC:=FRegister.R[4]; FRegister.R[4]:=B; end;
      $CD : begin B:=FRegister.ACC; FRegister.ACC:=FRegister.R[5]; FRegister.R[5]:=B; end;
      $CE : begin B:=FRegister.ACC; FRegister.ACC:=FRegister.R[6]; FRegister.R[6]:=B; end;
      $CF : begin B:=FRegister.ACC; FRegister.ACC:=FRegister.R[7]; FRegister.R[7]:=B; end;
            {pop dadr}
      $D0 : begin
              SetDataDirect(FCode[FRegister.PC+1],FData[FRegister.SP],False);
              B:=FRegister.SP; dec(B); FRegister.SP:=B; len:=2;
            end;
            {acall 6rel}
      $D1 : begin
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) mod 256;
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) div 256;
              FRegister.PC:=((FRegister.PC+2) and $E000)+6*256+FCode[FRegister.PC+1]; len:=0;
            end;
            {setb badr}
      $D2 : begin SetBit(FCode[FRegister.PC+1],True,False); len:=2; end;
            {setb c}
      $D3 : SetCarry(True);
            {da a}
      $D4 : FRegister.BCDCorrectAkku;
            {djnz dadr,rel}
      $D5 : begin
              B:=GetDataDirect(FCode[FRegister.PC+1],True); dec(B);
              SetDataDirect(FCode[FRegister.PC+1],B,True);
              If B<>0 then len:=3+ShortInt(FCode[FRegister.PC+2]) else len:=3;
            end;
            {xchd a,@ri}
      $D6 : begin
              B:=(FRegister.ACC and (16+32+64+128))+(GetDataIndirect(0) mod 16);
              C:=(GetDataIndirect(0) and (16+32+64+128))+(FRegister.ACC mod 16);
              FRegister.ACC:=B; SetDataIndirect(0,C);
            end;
      $D7 : begin
              B:=(FRegister.ACC and (16+32+64+128))+(GetDataIndirect(1) mod 16);
              C:=(GetDataIndirect(1) and (16+32+64+128))+(FRegister.ACC mod 16);
              FRegister.ACC:=B; SetDataIndirect(1,C);
            end;
      $D8 : begin
              B:=FRegister.R[0]; dec(B); FRegister.R[0]:=B;
              If B<>0 then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            end;
      $D9 : begin
              B:=FRegister.R[1]; dec(B); FRegister.R[1]:=B;
              If B<>0 then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            end;
      $DA : begin
              B:=FRegister.R[2]; dec(B); FRegister.R[2]:=B;
              If B<>0 then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            end;
      $DB : begin
              B:=FRegister.R[3]; dec(B); FRegister.R[3]:=B;
              If B<>0 then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            end;
      $DC : begin
              B:=FRegister.R[4]; dec(B); FRegister.R[4]:=B;
              If B<>0 then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            end;
      $DD : begin
              B:=FRegister.R[5]; dec(B); FRegister.R[5]:=B;
              If B<>0 then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            end;
      $DE : begin
              B:=FRegister.R[6]; dec(B); FRegister.R[6]:=B;
              If B<>0 then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            end;
      $DF : begin
              B:=FRegister.R[7]; dec(B); FRegister.R[7]:=B;
              If B<>0 then len:=2+ShortInt(FCode[FRegister.PC+1]) else len:=2;
            end;
            {movx a,@dptr}
      $E0 : FRegister.ACC:=GetDataExtern(FRegister.DPTR);
            {ajmp 7rel}
      $E1 : begin FRegister.PC:=((FRegister.PC+2) and $E000)+7*256+FCode[FRegister.PC+1]; len:=0; end;
            {movx a,@ri}
      $E2 : FRegister.ACC:=GetDataExternIndirect(0);
      $E3 : FRegister.ACC:=GetDataExternIndirect(1);
            {clr a}
      $E4 : FRegister.ACC:=0;
            {mov a,dadr}
      $E5 : begin FRegister.ACC:=GetDataDirect(FCode[FRegister.PC+1],False); len:=2; end;
            {mov a,@ri}
      $E6 : FRegister.ACC:=GetDataIndirect(0);
      $E7 : FRegister.ACC:=GetDataIndirect(1);
            {mov a,rx}
      $E8 : FRegister.ACC:=FRegister.R[0];
      $E9 : FRegister.ACC:=FRegister.R[1];
      $EA : FRegister.ACC:=FRegister.R[2];
      $EB : FRegister.ACC:=FRegister.R[3];
      $EC : FRegister.ACC:=FRegister.R[4];
      $ED : FRegister.ACC:=FRegister.R[5];
      $EE : FRegister.ACC:=FRegister.R[6];
      $EF : FRegister.ACC:=FRegister.R[7];
            {movx @dptr,a}
      $F0 : SetDataExtern(FRegister.DPTR,FRegister.ACC);
            {acall 7rel}
      $F1 : begin
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) mod 256;
              B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=(FRegister.PC+2) div 256;
              FRegister.PC:=((FRegister.PC+2) and $E000)+7*256+FCode[FRegister.PC+1]; len:=0;
            end;
            {movx @ri,a}
      $F2 : SetDataExternIndirect(0,FRegister.ACC);
      $F3 : SetDataExternIndirect(1,FRegister.ACC);
            {cpl a}
      $F4 : FRegister.ACC:=FRegister.ACC xor $FF;
            {mov dadr,a}
      $F5 : begin SetDataDirect(FCode[FRegister.PC+1],FRegister.ACC,False); len:=2; end;
            {mov @ri,a}
      $F6 : SetDataIndirect(0,FRegister.ACC);
      $F7 : SetDataIndirect(1,FRegister.ACC);
            {mov rx,a}
      $F8 : FRegister.R[0]:=FRegister.ACC;
      $F9 : FRegister.R[1]:=FRegister.ACC;
      $FA : FRegister.R[2]:=FRegister.ACC;
      $FB : FRegister.R[3]:=FRegister.ACC;
      $FC : FRegister.R[4]:=FRegister.ACC;
      $FD : FRegister.R[5]:=FRegister.ACC;
      $FE : FRegister.R[6]:=FRegister.ACC;
      $FF : FRegister.R[7]:=FRegister.ACC;
      else raise ESystemError.Create('Ungltiger Opcode "0x'+IntToHex(FCode[FRegister.PC],1)+'" an Adresse 0x'+IntToHex(FRegister.PC,1)+' ('+IntToStr(FRegister.PC)+').');
    end;
  finally
    inc(FRegister.PC,len);
  end;
end;

Procedure TProzessor.StartIRQ(Nr : Integer);
Var B : Byte;
begin
  FSFRArea[RegPCON]:=FSFRArea[RegPCON] and $9F;

  If Nr<>12 then FRegister.WaitingIRQ[Nr]:=False;
  B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=FRegister.PC mod 256;
  B:=FRegister.SP; inc(B); FRegister.SP:=B; FData[B]:=FRegister.PC div 256;

  If Nr<>12 then begin
    If FRegister.IRQLevel=1 then begin
      FRegister.IRQLevel:=2;
      FRegister.LowLovelBefore:=True;
    end else begin
      If Nr<6 then begin
        If FSFRArea[RegIP0] and (1 shl Nr) <>0 then FRegister.IRQLevel:=2 else FRegister.IRQLevel:=1;
      end else begin
        If FSFRArea[RegIP1] and (1 shl (Nr-6)) <>0 then FRegister.IRQLevel:=2 else FRegister.IRQLevel:=1;
      end;
      FRegister.LowLovelBefore:=False;
    end;
  end else FRegister.IRQLevel:=FRegister.IRQLevel+2;

  Case Nr of
    0..5  : FRegister.PC:=3+8*Nr;
    6..10 : FRegister.PC:=3+8*Nr+16;
    11    : FRegister.PC:=$6B;
    12    : FRegister.PC:=$7B;
  end;
end;

function TProzessor.Step : Boolean;
Var I : Integer;
    B : Byte;
begin
  result:=True;
  If not Assigned(FData) then raise ESystemError.Create('Datenspeicher nicht initialisiert.');
  If not Assigned(FCode) then raise ESystemError.Create('Codespeicher nicht initialisiert.');
  try

    SeriellWork;
    ExtIntWork;

    I:=FRegister.NextWaitingIRQ;
    If (I>=0) and (FCode[FRegister.PC]<>$32) then begin
      StartIRQ(I);
    end else begin

      If FSFRArea[RegPCON] and $40 = 0 then begin
        TimerWork;
        CapComWork;
        ADWork;
        B:=FCode[FRegister.PC];
        If (B=$84) or (B=$A4) then begin
          TimerWork; TimerWork; TimerWork;
          CapComWork; CapComWork; CapComWork;
          ADWork; ADWork; ADWork;
        end else begin
          For I:=0 to High(Long2Ops) do if B=Long2Ops[I] then begin
            TimerWork; CapComWork; ADWork; break;
          end;
        end;
      end;
      If (FSFRArea[RegPCON] and $60)<>0 then begin
        If (FSFRArea2[RegPCON1] and 128=128) and (FSFRArea[RegP3] and 4 =4) then StartIRQ(12);
        exit;
      end;
      FData.Addr:=FRegister.PC; FData.Protect:=True;

      try DoStep; finally FData.Protect:=False; FRegister.CalcParity; end;

    end;

  finally
    If (FRegister.PC=0) or (FRegister.PC=$918E) or (FRegister.PC=$91F9) then result:=False;
  end;
end;

function TProzessor.GetDataExtern(Addr: Word): Byte;
begin
  If FRegister.XMAP and (Addr>=$FF00) then result:=FXRAM[Addr-$FF00]
  else begin
    result:=FExt[Addr];
    FSFRArea[RegP0]:=$FF;
    FSFRArea[RegP2]:=Addr div 256;
  end;
end;

procedure TProzessor.SetDataExtern(Addr: Word; B: Byte);
begin
  If FRegister.XMAP and (Addr>=$FF00) then FXRAM[Addr-$FF00]:=B else begin
    FExt[Addr]:=B;
    If ((Addr<=$F700) and SingleMemory1) or ((Addr>$F700) and SingleMemory2) then FCode[Addr]:=B;
    FSFRArea[RegP0]:=$FF;
    FSFRArea[RegP2]:=Addr div 256;
  end;
end;

function TProzessor.GetDataExternIndirect(B: Byte): Byte;
begin
  If FRegister.XMAP then result:=FXRAM[FRegister.R[B]] else
  result:=GetDataExtern(FSFRArea[RegP2]*256+FRegister.R[B]);
end;

procedure TProzessor.SetDataExternIndirect(Addr, B: Byte);
begin
  If FRegister.XMAP then FXRAM[FRegister.R[Addr]]:=B else
  SetDataExtern(FSFRArea[RegP2]*256+FRegister.R[Addr],B);
end;

function TProzessor.GetBit(Nr: Byte; ReadModifyWrite : Boolean): Boolean;
Var B : Byte;
begin
  B:=1 shl (Nr mod 8);
  result:=(GetDataDirect(CodeToBit(Nr),ReadModifyWrite) and B = B);
end;

procedure TProzessor.SetBit(Nr: Byte; B: Boolean; ReadModifyWrite : Boolean);
Var C,A : Byte;
begin
  C:=1 shl (Nr mod 8); A:=CodeToBit(Nr);
  If B then
    SetDataDirect(A,GetDataDirect(A,ReadModifyWrite) or C,ReadModifyWrite)
  else
    SetDataDirect(A,GetDataDirect(A,ReadModifyWrite) and (255-C),ReadModifyWrite);
end;

function TProzessor.GetCarry: Boolean;
begin
  result:=(fCarry in FRegister.Flags);
end;

procedure TProzessor.SetCarry(B: Boolean);
begin
  If B then FRegister.PSW:=FRegister.PSW or 128 else FRegister.PSW:=FRegister.PSW and $7F;
end;

procedure TProzessor.TimerWork;
Var B,C,P3,H : Byte;
    Mode3 : Boolean;
begin
  {
  TCON: TF1 TR1 TF0 TR0 - - - -
  P3:    - - T1 T0 Int1 Int0 - -
  }
  B:=FSFRArea[RegTCON];
  C:=FSFRArea[RegTMOD];
  P3:=FSFRArea[RegP3];
  Mode3:=((C and (16+32)) div 16 = 3);
  If (B and 16 = 16) and ((P3 and 4 = 4) or (C and 8 = 0)) then begin
    If C and 4 = 4 then begin
      If T0Click or (LastT0 and (P3 and 16 = 0)) then IncTimer(0,C and 3,Mode3);
      LastT0:=(P3 and 16 = 16); T0Click:=False;
    end else IncTimer(0,C and 3,Mode3);
  end;
  If (B and 64 = 64) and ((P3 and 8 = 8) or (C and 128 = 0)) then begin
    If Mode3 then begin
      H:=FSFRArea[RegTH1];
      If H=255 then begin H:=0; FRegister.RequestIRQ(1); end else Inc(H);
      FSFRArea[RegTH1]:=H;
    end;
    If C and 64 = 64 then begin
      If T1Click or (LastT1 and (P3 and 32 = 0)) then IncTimer(1,(C and (16+32)) div 16,Mode3);
      LastT0:=(P3 and 32 = 32); T1Click:=False;
    end else IncTimer(1,(C and (16+32)) div 16,Mode3);
  end;
end;

procedure TProzessor.IncTimer(Nr, Mode : Byte; Mode3 : Boolean);
Var L,H : Byte;
begin
  L:=0; H:=0;
  Case Nr of
    0 : begin L:=FSFRArea[RegTL0]; H:=FSFRArea[RegTH0]; end;
    1 : begin L:=FSFRArea[RegTL1]; H:=FSFRArea[RegTH1]; end;
  end;
  Case Mode of
    0 : If H=255 then begin
          H:=0; If L mod 32 = 31 then begin
            L:=L xor 31; If (Nr=0) or (not Mode3) then FRegister.RequestIRQ(Nr*2+1);
          end else inc(L);
        end else inc(H);
    1 : If L=255 then begin
          L:=0; If H=255 then begin
            H:=0; If (Nr=0) or (not Mode3) then FRegister.RequestIRQ(Nr*2+1);
          end else inc(H);
        end else inc(L);
    2 : If L=255 then begin
          L:=H; If (Nr=0) or (not Mode3) then FRegister.RequestIRQ(Nr*2+1);
        end else Inc(L);
    3 : If Nr=0 then begin
          If L=255 then begin L:=0; FRegister.RequestIRQ(1); end else Inc(L);
        end;
  end;
  Case Nr of
    0 : begin FSFRArea[RegTL0]:=L; FSFRArea[RegTH0]:=H; end;
    1 : begin FSFRArea[RegTL1]:=L; FSFRArea[RegTH1]:=H; end;
  end;
end;

procedure TProzessor.SeriellWork;
begin
  inc(SeriellCounter);
  If (SeriellCounter mod FSeriellStep <>0) or (length(SeriellInBuffer)=0) then exit;
  FSFRArea.OnSBUFWrite:=nil;
  FSFRArea[RegSBUF]:=ord(SeriellInBuffer[1]);
  FSFRArea.OnSBUFWrite:=SendSBUF;
  SeriellInBuffer:=Copy(SeriellInBuffer,2,MaxInt);
  FSFRArea[RegSCON]:=FSFRArea[RegSCON] or 1;
  FRegister.RequestIRQ(4);
end;

Procedure TProzessor.SendSBUF(Sender : TObject);
begin
  If Assigned(FOnSeriellOutput) then FOnSeriellOutput(self,FSFRArea[RegSBUF]);
  FSFRArea[RegSCON]:=FSFRArea[RegSCON] or 2;
  FRegister.RequestIRQ(4);
end;

procedure TProzessor.ExtIntWork;
Var B,C,D,X : Byte;
begin
  B:=FSFRArea[RegP3];
  C:=FSFRArea[RegTCON];
  D:=FSFRArea[RegITCON];
  For X:=0 to 1 do begin
    If C and (1 shl (X*2)) = 0 then begin
      If (B and (4 shl X) = 0) and LastINT[X] then FRegister.RequestIRQ(X*2);
    end else begin
      If (B and (4 shl X) = 0) and LastINT[X] and (D and (2 shl (X*2))<>0) then FRegister.RequestIRQ(X*2);
      If (B and (4 shl X) <> 0) and not LastINT[X] and (D and (1 shl (X*2))<>0) then
        FRegister.RequestIRQ(X*2);
    end;
    LastINT[X]:=(B and (4 shl X) <> 0)
  end;
end;

procedure TProzessor.CapComWork;
Var B, CapComL, CapComH : Byte;
    Takte, Steps, Count, X : Integer;
    Running, Reset : Boolean;
begin
  inc(TaktCounter);

  B:=FSFRArea[RegCT1CON];

  Running:=(B and 8 = 8);
  Reset:=(B and 16 = 16);

  If Reset and ((not Running) or (Running and (not LastCapComRunning))) then begin
    FSFRArea[RegCT1OFL]:=0; FSFRArea[RegCT1OFH]:=0;
  end;
  If Running and (not LastCapComRunning) then RunningUp:=True;
  LastCapComRunning:=Running;

  If not Running then exit;
  Takte:=2 shl (B and 7);
  Steps:=0;

  Count:=0;
  While Count<12 do If Takte<20-Count then begin
    inc(Steps); inc(Count,Takte);
  end else begin
    If (TaktCounter mod 1000)<=(12*1000 div Takte) then begin
      inc(Steps); break;
    end;
  end;

  CapComL:=FSFRArea[RegCT1OFL]; CapComH:=FSFRArea[RegCT1OFH];
  For X:=1 to Steps do begin
    If B and 128 = 128 then begin
      If RunningUp then begin
        If CapComL=255 then begin
          If CapComH=255 then begin
            RunningUp:=False;
            FSFRArea[RegCCIR]:=FSFRArea[RegCCIR] and 128;
            FRegister.RequestIRQ(11);
          end else begin CapComL:=0; inc(CapComH); end;
        end else inc(CapComL);
        If (CapComL=FSFRArea[RegCCL0]) and (CapComH=FSFRArea[RegCCH0]) then SetCapComPort(0,True);
        If (CapComL=FSFRArea[RegCCL1]) and (CapComH=FSFRArea[RegCCH1]) then SetCapComPort(1,True);
        If (CapComL=FSFRArea[RegCCL2]) and (CapComH=FSFRArea[RegCCH2]) then SetCapComPort(2,True);
      end else begin
        If CapComL=0 then begin
          If CapComH=0 then begin
            CapComL:=FSFRArea[RegCCPL]; CapComH:=FSFRArea[RegCCPH];
            RunningUp:=True;
            FSFRArea[RegCCIR]:=FSFRArea[RegCCIR] and 64;
            FRegister.RequestIRQ(11);
          end else begin CapComL:=255; dec(CapComH); end;
        end else dec(CapComL);
        If (CapComL=FSFRArea[RegCCL0]) and (CapComH=FSFRArea[RegCCH0]) then SetCapComPort(0,False,True);
        If (CapComL=FSFRArea[RegCCL1]) and (CapComH=FSFRArea[RegCCH1]) then SetCapComPort(1,False,True);
        If (CapComL=FSFRArea[RegCCL2]) and (CapComH=FSFRArea[RegCCH2]) then SetCapComPort(2,False,True);
      end;
    end else begin
      If CapComL=255 then begin
        CapComL:=0;
        If CapComH=255 then begin
          CapComL:=FSFRArea[RegCCPL]; CapComH:=FSFRArea[RegCCPH];
          SetCapComPort(0,False); SetCapComPort(1,False); SetCapComPort(2,False);
          FSFRArea[RegCCIR]:=FSFRArea[RegCCIR] and 128;
          FRegister.RequestIRQ(11);
        end else inc(CapComH);
      end else inc(CapComL);
      If (CapComL=FSFRArea[RegCCL0]) and (CapComH=FSFRArea[RegCCH0]) then SetCapComPort(0,True);
      If (CapComL=FSFRArea[RegCCL1]) and (CapComH=FSFRArea[RegCCH1]) then SetCapComPort(1,True);
      If (CapComL=FSFRArea[RegCCL2]) and (CapComH=FSFRArea[RegCCH2]) then SetCapComPort(2,True);
    end;
  end;
  FSFRArea[RegCT1OFL]:=CapComL; FSFRArea[RegCT1OFH]:=CapComH;
end;

procedure TProzessor.SetCapComPort(Nr: Byte; Value: Boolean; CountingDown : Boolean = False);
Var B,P,Bit : Byte;
begin
  Case Nr of
    0 : B:=FSFRArea[RegCMSEL0] and 3;
    1 : B:=(FSFRArea[RegCMSEL0] and 48) div 16;
    2 : B:=FSFRArea[RegCMSEL1] and 3;
    else B:=0;
  end;
  P:=FSFRArea[RegP1];
  If B and 1 = 1 then begin
    Bit:=1 shl (2+(Nr*2));
    If Value then P:=P or Bit else P:=P and (255-Bit);
  end;
  If B and 2 = 2 then begin
    Bit:=1 shl (3+(Nr*2));
    If Value then P:=P or Bit else P:=P and (255-Bit);
  end;
  FSFRArea[RegP1]:=P;

  B:=B*2; If CountingDown then inc(B);
  FSFRArea[RegCCIR]:=FSFRArea[RegCCIR] or B;
  FRegister.RequestIRQ(10);
end;

procedure TProzessor.ADWork;
Var B,X,Nr : Byte;
    Value : Integer;
begin
  If TaktCounter mod 2<>0 then exit;
  If FSFRArea[RegADCON0] and 24 <> 0 then begin

    Value:=0;
    Nr:=FSFRArea[RegADCON0] mod 8;
    If Nr<=3 then begin
      If FSFRArea2[RegP1Ana] and (1 shl Nr) <>0 then begin
        If Assigned(FOnReadAnalog) then FOnReadAnalog(self,Nr,Value);
      end;
    end else begin
      If FSFRArea2[RegP1Ana] and (1 shl (Nr-2)) <>0 then begin
        If Assigned(FOnReadAnalog) then FOnReadAnalog(self,Nr,Value);
      end;
    end;
    {B:=FSFRArea2[RegP1Ana];
    Nr:=255;
    For X:=0 to 3 do If B and (1 shl X) <>0 then Nr:=X;
    If Nr<255 then begin
      Value:=0;
      If Assigned(FOnReadAnalog) then FOnReadAnalog(self,Nr,Value);
    end;
    B:=FSFRArea2[RegP3Ana]*4;
    Nr:=255;
    For X:=4 to 7 do If B and (1 shl X) <>0 then Nr:=X;
    If Nr<255 then begin
      Value:=0;
      If Assigned(FOnReadAnalog) then FOnReadAnalog(self,Nr,Value);
    end;}

    FSFRArea[RegADDATH]:=(Value mod $3FF) div 4;
    FSFRArea[RegADDATL]:=(Value mod 4)*64;
    If FSFRArea[RegADCON0] and 8 = 0 then
      FSFRArea[RegADCON0]:=FSFRArea[RegADCON0] and (255-16);
    FRegister.RequestIRQ(6);

  end;
end;

{ TRegister }

constructor TRegister.Create(const Data : TSpeicher; SFR, SFR2 : TSFRArea);
begin
  inherited Create;
  FData:=Data; FSFR:=SFR; FSFR2:=SFR2;
end;
procedure TRegister.Init(Addr : Integer);
Var X : Byte;
begin
  FSFR.Clear; FSFR2.Clear;
  ACC:=0; PC:=Addr; SP:=7;
  FSFR[RegP0]:=$FF; FSFR[RegP1]:=$FF;
  FSFR[RegP2]:=$FF; FSFR[RegP3]:=$FF;
  For X:=0 to 11 do WaitingIRQ[X]:=False;
  IRQLevel:=0; LowLovelBefore:=False;
  SeriellIRQ:=False; CapComIRQ1:=False; CapComIRQ2:=False;
  FSFR[RegSCON]:=FSFR[RegSCON] or 2;
end;

procedure TRegister.SetACC(I : Integer);
begin
  If I>0 then begin
    If I<256 then Flags:=Flags-[fCarry] else Flags:=Flags+[fCarry];
      ACC:=I mod 256;
  end else begin
    If I<0 then begin Flags:=Flags+[fCarry]; inc(I,256); end else Flags:=Flags-[fCarry];
    ACC:=I;
  end;
end;

function TRegister.GetR(I: Integer): Byte;
Var Start : Integer;
    F : Byte;
begin
  Start:=0; F:=PSW;
  If (F and 8) = 8 then inc(Start,8);
  If (F and 16) = 16 then inc(Start,16);
  result:=FData[Start+I];
end;

procedure TRegister.SetR(I: Integer; B: Byte);
Var Start : Integer;
    F : Byte;
begin
  Start:=0; F:=PSW;
  If (F and 8) = 8 then inc(Start,8);
  If (F and 16) = 16 then inc(Start,16);
  FData[Start+I]:=B;
end;

procedure TRegister.SetFlags(F: TFlagSet);
Var B : Byte;
begin
  B:=0;
  If fParity in F then inc(B,1);
  If fF1 in F then inc(B,2);
  If fOverflow in F then inc(B,4);
  If fRS0 in F then inc(B,8);
  If fRS1 in F then inc(B,16);
  If fF0 in F then inc(B,32);
  If fAux in F then inc(B,64);
  If fCarry in F then inc(B,128);
  PSW:=B;
end;

function TRegister.GetFlags: TFlagSet;
Var B : Byte;
begin
  B:=PSW;
  result:=[];
  If B and   1 =   1 then result:=result+[fParity];
  If B and   2 =   2 then result:=result+[fF1];
  If B and   4 =   4 then result:=result+[fOverflow];
  If B and   8 =   8 then result:=result+[fRS0];
  If B and  16 =  16 then result:=result+[fRS1];
  If B and  32 =  32 then result:=result+[fF0];
  If B and  64 =  64 then result:=result+[fAux];
  If B and 128 = 128 then result:=result+[fCarry];
end;

procedure TRegister.AddToACC(B : Byte; WithCarry : Boolean);
Var Old,Neu : Byte;
    F : Word;
    Fl : TFlagSet;
begin
  Old:=Acc; If WithCarry and (fCarry in Flags) then F:=1 else F:=0;
  asm
    mov ax,F
    push ax
    popf

    mov ah,Old
    adc ah,B
    mov Neu,ah

    pushf
    pop ax
    mov F,ax
  end;
  ACC:=Neu; Fl:=GetFlags;
  If F and    1 =    1 then Fl:=Fl+[fCarry] else Fl:=Fl-[fCarry];
  If F and   16 =   16 then Fl:=Fl+[fAux] else Fl:=Fl-[fAux];
  If F and 2048 = 2048 then Fl:=Fl+[fOverflow] else Fl:=Fl-[fOverflow];
  SetFlags(Fl);
end;

procedure TRegister.SubFromACC(B : Byte);
Var Old,Neu : Byte;
    F : Word;
    Fl : TFlagSet;
begin
  Old:=Acc; If fCarry in Flags then F:=1 else F:=0;
  asm
    mov ax,F
    push ax
    popf

    mov ah,Old
    sbb ah,B
    mov Neu,ah

    pushf
    pop ax
    mov F,ax
  end;
  ACC:=Neu; Fl:=GetFlags;
  If F and    1 =    1 then Fl:=Fl+[fCarry] else Fl:=Fl-[fCarry];
  If F and   16 =   16 then Fl:=Fl+[fAux] else Fl:=Fl-[fAux];
  If F and 2048 = 2048 then Fl:=Fl+[fOverflow] else Fl:=Fl-[fOverflow];
  SetFlags(Fl);
end;

function TRegister.GetByte(Index: Integer): Byte;
begin
  result:=0;
  Case Index of
    1 : result:=FSFR[RegA];
    2 : result:=FSFR[RegB];
    3 : result:=FSFR[RegPSW];
    4 : result:=FSFR[RegSP];
  end;
end;

procedure TRegister.SetByte(Index: Integer; B: Byte);
begin
  Case Index of
    1 : FSFR[RegA]:=B;
    2 : FSFR[RegB]:=B;
    3 : FSFR[RegPSW]:=B;
    4 : FSFR[RegSP]:=B;
  end;
end;

function TRegister.GetWord(Index: Integer): Word;
begin
  result:=0;
  Case Index of
    1 : result:=FSFR[RegDPL]+256*FSFR[RegDPH];
  end;
end;

procedure TRegister.SetWord(Index: Integer; W: Word);
begin
  Case Index of
    1 : begin FSFR[RegDPL]:=W mod 256; FSFR[RegDPH]:=W div 256; end;
  end;
end;

procedure TRegister.BCDCorrectAkku;
Var A : Byte;
    F : Word;
begin
  A:=ACC;
  If fCarry in Flags then F:=1 else F:=0;
  If fAux in Flags then inc(F,16);
  asm
    mov ax,F
    push ax
    popf

    mov al,A
    daa
    mov A,al
  end;
end;

function TRegister.GetSFRBit(Index: Integer): Boolean;
begin
  result:=False;
  Case Index of
    1 : result:=(FSFR[RegSYSCON] and 1 = 1);
    2 : result:=(FSFR[RegSYSCON] and 16 = 16);

  end;
end;

procedure TRegister.SetSFRBit(Index: Integer; B: Boolean);
begin
  Case Index of
    1 : If B then FSFR[RegSYSCON]:=FSFR[RegSYSCON] or 1
             else FSFR[RegSYSCON]:=FSFR[RegSYSCON] and (255-1);
    2 : If B then FSFR[RegSYSCON]:=FSFR[RegSYSCON] or 16
             else FSFR[RegSYSCON]:=FSFR[RegSYSCON] and (255-16);
  end;
end;

{
  TCON:     TF1     -   TF0     -   IE1     -   IE0     -
  SCON:       -     -     -     -     -     -    RI    TI
  T2CON:    TF2  EXF2     -     -     -     -     -     -
  ADCON0:     -     -  IADC     -     -     -     -     -
  ITCON:      -   IE2     -     -     -     -     -     -
  TRCON:      -   TRF     -     -     -     -     -     -
  BCON:       -     -     -     - BCERR     -     -     -
  CT2CON:  CT2P     -     -     -     -     -     -     -
  CCIR    CT1FP CT1FC  CC2F  CC2R  CC1F  CC1R  CC0F  CC0R

  IE0                  =  1. Int
  TF0                  =  2. Int
  IE1                  =  3. Int
  TF1                  =  4. Int
  RI+TI                =  5. Int
  TF2+EXF2             =  6. Int
  IADC                 =  7. Int
  IE2                  =  8. Int
  TRF+BCERR            =  9. Int
  CT2P                 = 10. Int
  CC0F-CC2F, CC0R-CC2R = 11. Int
  CT1FP, CT1FC         = 12. Int
}

function TRegister.GetWaitingIRQ(I: Integer): Boolean;
begin
  result:=False;
  If (I<0) or (I>11) then exit;
  Case I of
     0 : result:=FSFR[RegTCON] and 2 = 2;
     1 : result:=FSFR[RegTCON] and 32 = 32;
     2 : result:=FSFR[RegTCON] and 8 = 8;
     3 : result:=FSFR[RegTCON] and 128 = 128;
     4 : result:=SeriellIRQ;
     5 : result:=FSFR[RegT2CON] and (128+64) <> 0;
     6 : result:=FSFR[RegADCON0] and 32 = 32;
     7 : result:=FSFR[RegITCON] and 64 = 64;
     8 : result:=(FSFR[RegTRCON] and 64 = 64) or (FSFR[RegBCON] and 8 = 8);
     9 : result:=FSFR[RegCT2CON] and 128 = 128;
    10 : result:=CapComIRQ1;
    11 : result:=CapComIRQ2;
  end;
end;

procedure TRegister.SetWaitingIRQ(I: Integer; Value: Boolean);
begin
  If (I<0) or (I>11) then exit;
  If Value then Case I of
     0 : FSFR[RegTCON]:=FSFR[RegTCON] or 2;
     1 : FSFR[RegTCON]:=FSFR[RegTCON] or 32;
     2 : FSFR[RegTCON]:=FSFR[RegTCON] or 8;
     3 : FSFR[RegTCON]:=FSFR[RegTCON] or 128;
     4 : SeriellIRQ:=True;
     5 : FSFR[RegT2CON]:=FSFR[RegT2CON] or (128+64);
     6 : FSFR[RegADCON0]:=FSFR[RegADCON0] or 32;
     7 : FSFR[RegITCON]:=FSFR[RegITCON] or 64;
     8 : begin FSFR[RegTRCON]:=FSFR[RegTRCON] or 64; FSFR[RegBCON]:=FSFR[RegBCON] or 8; end;
     9 : FSFR[RegCT2CON]:=FSFR[RegCT2CON] or 128;
    10 : CapComIRQ1:=True;
    11 : CapComIRQ2:=True;
  end else Case I of
     0 : FSFR[RegTCON]:=FSFR[RegTCON] and (255-2);
     1 : FSFR[RegTCON]:=FSFR[RegTCON] and (255-32);
     2 : FSFR[RegTCON]:=FSFR[RegTCON] and (255-8);
     3 : FSFR[RegTCON]:=FSFR[RegTCON] and (255-128);
     4 : SeriellIRQ:=False;
     5 : FSFR[RegT2CON]:=FSFR[RegT2CON] and (255-128-64);
     6 : FSFR[RegADCON0]:=FSFR[RegADCON0] and (255-32);
     7 : FSFR[RegITCON]:=FSFR[RegITCON] and (255-64);
     8 : begin FSFR[RegTRCON]:=FSFR[RegTRCON] and (256-64); FSFR[RegBCON]:=FSFR[RegBCON] and (256-8); end;
     9 : FSFR[RegCT2CON]:=FSFR[RegCT2CON] and (256-128);
    10 : CapComIRQ1:=False;
    11 : CapComIRQ2:=False;
  end;
end;

Procedure TRegister.RequestIRQ(Nr : Byte);
Var B,C : Byte;
begin
  If Nr>11 then exit;
  B:=FSFR[RegIEN0]; If B<128 then exit;
  If Nr>5 then begin B:=FSFR[RegIEN1]; C:=1 shl (Nr-6); end else C:=1 shl Nr;
  If B and C = 0 then exit;
  SetWaitingIRQ(Nr,True);
end;

function TRegister.NextWaitingIRQ: Integer;
Var P,X : Byte;
begin
  result:=-1;
  If IRQLevel=2 then exit;
  P:=FSFR[RegIP0];
  For X:=0 to 5 do If GetWaitingIRQ(X) then If (IRQLevel=0) or (P and (1 shl X) <>0) then begin
    result:=X; exit;
  end;
  P:=FSFR[RegIP1];
  For X:=0 to 5 do If GetWaitingIRQ(X+6) then If (IRQLevel=0) or (P and (1 shl X) <>0) then begin
    result:=X+6; exit;
  end;
end;

procedure TRegister.CalcParity;
Var B,I : Byte;
begin
  B:=ACC; I:=0;
  while B<>0 do begin If B mod 2 = 1 then inc(I); B:=B div 2; end;
  If I mod 2 = 1
    then Flags:=Flags+[fParity]
    else Flags:=Flags-[fParity];
end;

end.
