unit Compiler;
interface

uses Classes, Tools;

Type TProgramm=class;

     TCommand=class
  private
    TempBefehl, TempParameter : String;
    FAddr : Integer;
    FZeile : Integer;
    FCode : String;
    FOwner : TProgramm;
    Procedure Parts(var S,T : String);
    Function FindTarget(Var I : Integer) : Boolean;
    Function RelDest(S : String; var B : Byte; len : Byte = 2) : Boolean; overload;
    Function RelDest(var B : Byte; len : Byte = 2) : Boolean; overload;
    Function RelDest2(var B,C : Byte; len : Byte = 2) : Boolean; overload;
  public
    AsmCode : Array of Byte;
    Constructor Create(AOwner : TProgramm);
    Function VarTest(S : String; var I : Integer; PreChar : Char = #0) : Boolean;
    Procedure SetCode(A : Array of Byte; AddAddr : Boolean = True);
    Procedure Compile(S : String; Nr : Integer);
    Procedure AddWords(S : String; Nr : Integer);
    Procedure AddBytes(S : String; Nr : Integer);
    Procedure SolveReferenzes;
  published
    property Zeile : Integer read FZeile;
    property Addr : Integer read FAddr;
    property Code : String read FCode;
end;

     TProgramm=class
  private
    FVarTab, FTargetTab : TStrings;
    FRAMAddr, FRAMValue : TList;
    FProtAddr : TList;
    FStartAddr, FNextAddr : Integer;
    FRAMFile : String;
    FTakte : Integer;
    FClearRAM, FProtectDefault : Boolean;
    FCommand : Array of TCommand;
    Procedure SetVariable(Nr : Integer; S : String);
    Function ScanLine(Nr : Integer; S : String) : Boolean;
    Function CompileLine(Nr : Integer; S : String) : Boolean;
    Procedure FreeCommands;
    Function GetCommandCount : Integer;
    Function GetCommand(I : Integer) : TCommand;
    Function LoadRamVar(S : String; Nr : Integer; Prot : Boolean) : Boolean;
    Procedure ProtectDefault;
    Procedure Parts(var S, T: String; Nr : Integer);
  public
    Constructor Create;
    Destructor Destroy; override;
    Procedure Compile(const Lines : TStrings; MemoryProtect : Boolean);
    Function GetAsmCode(MiniOS : Boolean) : TStream;
    Function GetAsmCode2 : TStrings;
    Function GetListing : TStrings;
    property StartAddr : Integer read FStartAddr;
    property CommandCount : Integer read GetCommandCount;
    property Command[I : Integer] : TCommand read GetCommand;
    property RAMFile : String read FRAMFile;
    property ClearRAM : Boolean read FClearRAM;
    property RAMAddr : TList read FRAMAddr;
    property RAMValue : TList read FRAMValue;
    property ProtectAddr : TList read FProtAddr;
    property Takte : Integer read FTakte;
end;

implementation

uses SysUtils, PrgConst;

{ TCommand }

constructor TCommand.Create(AOwner: TProgramm);
begin
  inherited Create;
  FOwner:=AOwner;
  SetLength(AsmCode,0); FAddr:=0; FCode:='';
end;

Function TCommand.VarTest(S : String; var I : Integer; PreChar : Char) : Boolean;
Var X : Integer;
begin
  result:=False;
  If PreChar<>#0 then begin
    If (S='') or (S[1]<>PreChar) then exit;
    S:=Trim(Copy(S,2,MaxInt));
  end;
  If IsByteZahl(S) then begin
    result:=True; I:=StrToInt(S);
  end else begin
    result:=False;
    For X:=0 to FOwner.FVarTab.Count-1 do If S=FOwner.FVarTab[X] then begin
      I:=Integer(FOwner.FVarTab.Objects[X]); result:=True; exit;
    end;
    For X:=0 to AllTimeTab.Count-1 do If S=AllTimeTab[X] then begin
      I:=Integer(AllTimeTab.Objects[X]); result:=True; exit;
    end;
  end;
end;

Procedure TCommand.SetCode(A : Array of Byte; AddAddr : Boolean);
Var X : Integer;
begin
  SetLength(AsmCode,High(A)+1);
  For X:=0 to High(A) do AsmCode[X]:=A[X];
  If AddAddr then inc(FOwner.FNextAddr,length(AsmCode));
end;

procedure TCommand.Compile(S: String; Nr: Integer);
Var T,U : String;
    I,J,L : Integer;
    B : Byte;
    B1,B2,B4 : Boolean;
begin
  FCode:=S; FAddr:=FOwner.FNextAddr; FZeile:=Nr;

  {nop} If S='nop' then begin SetCode([$00]); exit; end;

  {mov}
  If Left('mov ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); Parts(S,T);
    B1:=VarTest(S,I); B2:=VarTest(T,J); B4:=VarTest(T,L,'#');
    {mov addr,a} If B1 and (T='a') then begin SetCode([$F5,I]); exit; end;
    {mov a,addr} If (S='a') and B2 then begin SetCode([$E5,J]); exit; end;
    {mov a,rx} If (S='a') and IsRx(T) then begin SetCode([$E8+ord(T[2])-48]); exit; end;
    {mov rx,a} If IsRx(S) and (T='a') then begin SetCode([$F8+ord(S[2])-48]); exit; end;
    {mov a,#const8} If (S='a') and B4 then begin SetCode([$74,L]); exit; end;
    {mov rx,#const8} If IsRx(S) and B4 then begin SetCode([$78+ord(S[2])-48,L]); exit; end;
    {mov addr,#const8} If B1 and B4 then begin SetCode([$75,I,L]); exit; end;
    {mov rx,addr} If IsRx(S) and B2 then begin SetCode([$A8+ord(S[2])-48,J]); exit; end;
    {mov addr,rx} If B1 and IsRx(T) then begin SetCode([$88+ord(T[2])-48,I]); exit; end;
    {mov addr,addr} If B1 and B2 then begin SetCode([$85,I,J]); exit; end;
    {mov DPTR,#const16} If S='dptr' then begin
                          If IsWordZahl(T,'#') then begin
                            I:=StrToInt(Copy(T,2,MaxInt)); SetCode([$90,I div 256,I mod 256]);
                          end else begin
                            TempBefehl:='mov'; TempParameter:='dptr,'+T; inc(FOwner.FNextAddr,3);
                          end;
                          exit;
                        end;
    {mov a,@r0} If (S='a') and (T='@r0') then begin SetCode([$E6]); exit; end;
    {mov a,@r1} If (S='a') and (T='@r1') then begin SetCode([$E7]); exit; end;
    {mov @r0,a} If (S='@r0') and (T='a') then begin SetCode([$F6]); exit; end;
    {mov @r1,a} If (S='@r1') and (T='a') then begin SetCode([$F7]); exit; end;
    {mov @r0,#const8} If (S='@r0') and B4 then begin SetCode([$76,L]); exit; end;
    {mov @r1,#const8} If (S='@r1') and B4 then begin SetCode([$77,L]); exit; end;
    {mov dadr,@r0} If B1 and (T='@r0') then begin SetCode([$86,I]); exit; end;
    {mov dadr,@r1} If B1 and (T='@r1') then begin SetCode([$87,I]); exit; end;
    {mov @r0,dadr} If (S='@r0') and B2 then begin SetCode([$A6,J]); exit; end;
    {mov @r1,dadr} If (S='@r1') and B2 then begin SetCode([$A7,J]); exit; end;
    {mov c,badr} If (S='c') and IsBitAddr(T) then begin SetCode([$A2,GetBitAddr(T)]); exit; end;
    {mov badr,c} If IsBitAddr(S) and (T='c') then begin SetCode([$92,GetBitAddr(T)]); exit; end;
  end;

  {movx}
  If Left('movx ',S) then begin
    S:=Trim(Copy(S,6,MaxInt)); Parts(S,T);
    {movx a,@r0} If (S='a') and (T='@r0') then begin SetCode([$E2]); exit; end;
    {movx a,@r1} If (S='a') and (T='@r1') then begin SetCode([$E3]); exit; end;
    {movx @r0,a} If (S='@r0') and (T='a') then begin SetCode([$F2]); exit; end;
    {movx @r1,a} If (S='@r1') and (T='a') then begin SetCode([$F3]); exit; end;
    {movx a,@dptr} If (S='a') and (T='@dptr') then begin SetCode([$E0]); exit; end;
    {movx @dptr,a} If (S='@dptr') and (T='a') then begin SetCode([$F0]); exit; end;
  end;

  {movc}
  If Left('movc ',S) then begin
    S:=Trim(Copy(S,6,MaxInt)); Parts(S,T); If S='a' then begin
      {movc a,@a+dptr} If T='@a+dptr' then begin SetCode([$93]); exit; end;
      {movc a,@a+pc} If T='@a+pc' then begin SetCode([$83]); exit; end;
    end;
  end;

  {push}
  If Left('push ',S) then begin
    S:=Trim(Copy(S,6,MaxInt)); B1:=VarTest(S,I);
    If B1 then begin SetCode([$C0,I]); exit; end;
  end;

  {pop}
  If Left('pop ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); B1:=VarTest(S,I);
    If B1 then begin SetCode([$D0,I]); exit; end;
  end;

  {add}
  If Left('add ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); Parts(S,T); B2:=VarTest(T,J); B4:=VarTest(T,L,'#'); If S='a' then begin
      {add a,#const8} If B4 then begin SetCode([$24,L]); exit; end;
      {add a,dadr} If B2 then begin SetCode([$25,J]); exit; end;
      {add a,rx} If IsRx(T) then begin SetCode([$28+ord(T[2])-48]); exit; end;
      {add a,@r0} If T='@r0' then begin SetCode([$26]); exit; end;
      {add a,@r1} If T='@r1' then begin SetCode([$27]); exit; end;
    end;
  end;

  {addc}
  If Left('addc ',S) then begin
    S:=Trim(Copy(S,6,MaxInt)); Parts(S,T); B2:=VarTest(T,J); B4:=VarTest(T,L,'#'); If S='a' then begin
      {addc a,#const8} If B4 then begin SetCode([$34,L]); exit; end;
      {addc a,dadr} If B2 then begin SetCode([$35,J]); exit; end;
      {addc a,rx} If IsRx(T) then begin SetCode([$38+ord(T[2])-48]); exit; end;
      {addc a,@r0} If T='@r0' then begin SetCode([$36]); exit; end;
      {addc a,@r1} If T='@r1' then begin SetCode([$37]); exit; end;
    end;
  end;

  {xch}
  If Left('xch ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); Parts(S,T); B2:=VarTest(T,J); If S='a' then begin
      {xch a,rx} If IsRx(T) then begin SetCode([$C8+ord(S[2])-48]); exit; end;
      {xch a,dadr} If B2 then begin SetCode([$C5,J]); exit; end;
      {xch a,@r0} If T='@r0' then begin SetCode([$C6]); exit; end;
      {xch a,@r1} If T='@r1' then begin SetCode([$C7]); exit; end;
    end;
  end;

  {xchd}
  If Left('xchd ',S) then begin
    S:=Trim(Copy(S,6,MaxInt)); Parts(S,T); If S='a' then begin
      {xchd a,@r0} If T='@r0' then begin SetCode([$D6]); exit; end;
      {xchd a,@r1} If T='@r1' then begin SetCode([$D7]); exit; end;
    end;
  end;

  {swap}
  If S='swap a' then begin SetCode([$C4]); exit; end;

  {ljmp}
  If Left('ljmp ',S) then begin
    S:=Trim(Copy(S,6,MaxInt));
    If IsWordZahl(S) then begin I:=StrToInt(S); SetCode([$02,I div 256,I mod 256]);
    end else begin inc(FOwner.FNextAddr,3); TempBefehl:='ljmp'; TempParameter:=S; end;
    exit;
  end;

  {lcall}
  If Left('lcall ',S) then begin
    S:=Trim(Copy(S,7,MaxInt));
    If IsWordZahl(S) then begin I:=StrToInt(S); SetCode([$12,I div 256,I mod 256]);
    end else begin inc(FOwner.FNextAddr,3); TempBefehl:='lcall'; TempParameter:=S; end;
    exit;
  end;

  {inc}
  If Left('inc ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); B1:=VarTest(S,I);
    {inc a} If S='a' then begin SetCode([$04]); exit; end;
    {inc rx} If IsRx(S) then begin SetCode([$08+ord(S[2])-48]); exit; end;
    {inc dadr} If B1 then begin SetCode([$05,I]); exit; end;
    {inc dptr} If S='dptr' then begin SetCode([$A3]); exit; end;
    {inc @r0} If S='@r0' then begin SetCode([$06]); exit; end;
    {inc @r1} If S='@r1' then begin SetCode([$07]); exit; end;
  end;

  {subb}
  If Left('subb ',S) then begin
    S:=Trim(Copy(S,6,MaxInt)); Parts(S,T); B2:=VarTest(T,J); B4:=VarTest(T,L,'#'); If S='a' then begin
      {subb a,rx} If IsRx(T) then begin SetCode([$98+ord(T[2])-48]); exit; end;
      {subb a,dadr} If B2 then begin SetCode([$95,J]); exit; end;
      {subb a,#const8} If B4 then begin SetCode([$94,L]); exit; end;
      {subb a,@r0} If T='@r0' then begin SetCode([$96]); exit; end;
      {subb a,@r1} If T='@r1' then begin SetCode([$97]); exit; end;
    end;
  end;

  {dec}
  If Left('dec ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); B1:=VarTest(S,I);
    {dec a} If S='a' then begin SetCode([$14]); exit; end;
    {dec rx} If IsRx(S) then begin SetCode([$18+ord(S[2])-48]); exit; end;
    {dec dadr} If B1 then begin SetCode([$15,I]); exit; end;
    {dec @r0} If S='@r0' then begin SetCode([$16]); exit; end;
    {dec @r1} If S='@r1' then begin SetCode([$17]); exit; end;
  end;

  {mul ab}
  If S='mul ab' then begin SetCode([$A4]); exit; end;

  {div ab}
  If S='div ab' then begin SetCode([$84]); exit; end;

  {da a}
  If S='da a' then begin SetCode([$D4]); exit; end;

  {anl}
  If Left('anl ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); Parts(S,T); B1:=VarTest(S,I); B2:=VarTest(T,J); B4:=VarTest(T,L,'#');
    {anl a,rx} If (S='a') and IsRx(T) then begin SetCode([$58+ord(T[2])-48]); exit; end;
    {anl a,@r0} If (S='a') and (T='@r0') then begin SetCode([$56]); exit; end;
    {anl a,@r1} If (S='a') and (T='@r1') then begin SetCode([$57]); exit; end;
    {anl a,dadr} If (S='a') and B2 then begin SetCode([$55,J]); exit; end;
    {anl a,#const8} If (S='a') and B4 then begin SetCode([$54,L]); exit; end;
    {anl dadr,a} If B1 and (T='a') then begin SetCode([$52,I]); exit; end;
    {anl dadr,#const8} If B1 and B4 then begin SetCode([$53,I,L]); exit; end;
    {anl c,badr} If (S='c') and IsBitAddr(T) then begin SetCode([$82,GetBitAddr(T)]); exit; end;
    {anl c,/badr} If (S='c') and (length(T)>1) and (T[1]='/') and IsBitAddr(Trim(Copy(T,2,MaxInt))) then begin SetCode([$B0,GetBitAddr(Trim(Copy(T,2,MaxInt)))]); exit; end;
  end;

  {orl}
  If Left('orl ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); Parts(S,T); B1:=VarTest(S,I); B2:=VarTest(T,J); B4:=VarTest(T,L,'#');
    {orl a,rx} If (S='a') and IsRx(T) then begin SetCode([$48+ord(T[2])-48]); exit; end;
    {orl a,@r0} If (S='a') and (T='@r0') then begin SetCode([$46]); exit; end;
    {orl a,@r1} If (S='a') and (T='@r1') then begin SetCode([$47]); exit; end;
    {orl a,dadr} If (S='a') and B2 then begin SetCode([$45,J]); exit; end;
    {orl a,#const8} If (S='a') and B4 then begin SetCode([$44,L]); exit; end;
    {orl dadr,a} If B1 and (T='a') then begin SetCode([$42,I]); exit; end;
    {orl dadr,#const8} If B1 and B4 then begin SetCode([$43,I,L]); exit; end;
    {orl c,badr} If (S='c') and IsBitAddr(T) then begin SetCode([$72,GetBitAddr(T)]); exit; end;
    {orl c,/badr} If (S='c') and (length(T)>1) and (T[1]='/') and IsBitAddr(Trim(Copy(T,2,MaxInt))) then begin SetCode([$A0,GetBitAddr(Trim(Copy(T,2,MaxInt)))]); exit; end;
  end;

  {xrl}
  If Left('xrl ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); Parts(S,T); B1:=VarTest(S,I); B2:=VarTest(T,J); B4:=VarTest(T,L,'#');
    {xrl a,rx} If (S='a') and IsRx(T) then begin SetCode([$68+ord(T[2])-48]); exit; end;
    {xrl a,@r0} If (S='a') and (T='@r0') then begin SetCode([$66]); exit; end;
    {xrl a,@r1} If (S='a') and (T='@r1') then begin SetCode([$67]); exit; end;
    {xrl a,dadr} If (S='a') and B2 then begin SetCode([$65,J]); exit; end;
    {xrl a,#const8} If (S='a') and B4 then begin SetCode([$64,L]); exit; end;
    {xrl dadr,a} If B1 and (T='a') then begin SetCode([$62,I]); exit; end;
    {xrl dadr,#const8} If B1 and B4 then begin SetCode([$63,I,L]); exit; end;
  end;

  {cpl}
  If Left('cpl ',S) then begin
    S:=Trim(Copy(S,5,MaxInt));
    {cpl a} If S='a' then begin SetCode([$F4]); exit; end;
    {cpl c} If S='c' then begin SetCode([$B3]); exit; end;
    {cpl badr} If IsBitAddr(S) then begin SetCode([$B2,GetBitAddr(S)]); exit; end;
  end;

  {clr}
  If Left('clr ',S) then begin
    S:=Trim(Copy(S,5,MaxInt));
    {clr a} If S='a' then begin SetCode([$E4]); exit; end;
    {clr c} If S='c' then begin SetCode([$C3]); exit; end;
    {clr badr} If IsBitAddr(S) then begin SetCode([$C2,GetBitAddr(S)]); exit; end;
  end;

  {setb}
  If Left('setb ',S) then begin
    S:=Trim(Copy(S,6,MaxInt));
    {setb c} If S='c' then begin SetCode([$D3]); exit; end;
    {setb badr} If IsBitAddr(S) then begin SetCode([$D2,GetBitAddr(S)]); exit; end;
  end;

  {rl a} If S='rl a' then begin SetCode([$23]); exit; end;
  {rlc a} If S='rlc a' then begin SetCode([$33]); exit; end;
  {rr a} If S='rr a' then begin SetCode([$03]); exit; end;
  {rrc a} If S='rrc a' then begin SetCode([$13]); exit; end;

  {ret} If S='ret' then begin SetCode([$22]); exit; end;
  {reti} If S='reti' then begin SetCode([$32]); exit; end;

  {jmp @a+dptr} If S='jmp @a+dptr' then begin SetCode([$73]); exit; end;

  {sjmp rel}
  If Left('sjmp ',S) then begin
    S:=Trim(Copy(S,6,MaxInt));
    If RelDest(S,B) then SetCode([$80,B]) else begin
      inc(FOwner.FNextAddr,2); TempBefehl:='sjmp'; TempParameter:=S;
    end;
    exit;
  end;

  {jc rel}
  If Left('jc ',S) then begin
    S:=Trim(Copy(S,4,MaxInt));
    If RelDest(S,B) then SetCode([$40,B]) else begin
      inc(FOwner.FNextAddr,2); TempBefehl:='jc'; TempParameter:=S;
    end;
    exit;
  end;

  {jnc rel}
  If Left('jnc ',S) then begin
    S:=Trim(Copy(S,5,MaxInt));
    If RelDest(S,B) then SetCode([$50,B]) else begin
      inc(FOwner.FNextAddr,2); TempBefehl:='jnc'; TempParameter:=S;
    end;
    exit;
  end;

  {jz rel}
  If Left('jz ',S) then begin
    S:=Trim(Copy(S,4,MaxInt));
    If RelDest(S,B) then SetCode([$60,B]) else begin
      inc(FOwner.FNextAddr,2); TempBefehl:='jz'; TempParameter:=S;
    end;
    exit;
  end;

  {jnz rel}
  If Left('jnz ',S) then begin
    S:=Trim(Copy(S,5,MaxInt));
    If RelDest(S,B) then SetCode([$70,B]) else begin
      inc(FOwner.FNextAddr,2); TempBefehl:='jnz'; TempParameter:=S;
    end;
    exit;
  end;

  {djnz}
  If Left('djnz ',S) then begin
    S:=Trim(Copy(S,6,MaxInt)); Parts(S,T); B1:=VarTest(S,I);
    {djnz dadr,rel} If B1 then begin
                      If RelDest(T,B) then SetCode([$D5,I,B]) else begin
                        inc(FOwner.FNextAddr,3); TempBefehl:='djnz'; TempParameter:=S+','+T;
                      end;
                      exit;
                    end;
    {djnz Re,rel} If IsRx(S) then begin
                    If RelDest(T,B) then SetCode([$D8+ord(S[2])-48,B]) else begin
                      inc(FOwner.FNextAddr,2); TempBefehl:='djnz'; TempParameter:=S+','+T;
                    end;
                    exit;
                  end;
  end;

  {cjne}
  If Left('cjne ',S) then begin
    S:=Trim(Copy(S,6,MaxInt)); Parts(S,T); Parts(T,U); B4:=VarTest(T,L,'#');
    If RelDest(U,B,3) then begin
      {cjne a,dadr,rel} If (S='a') and VarTest(T,I) then begin SetCode([$B5,I,B]); exit; end;
      {cjne a,#const8,rel} If (S='a') and B4 then begin SetCode([$B4,L,B]); exit; end;
      {cjne Rx,#const8,rel} If IsRx(S) and B4 then begin SetCode([$B8+ord(S[2])-48,L,B]); exit; end;
      {cjne @r0,#const8,rel} If (S='@r0') and B4 then begin SetCode([$B6,L,B]); exit; end;
      {cjne @r1,#const8,rel} If (S='@r1') and B4 then begin SetCode([$B7,L,B]); exit; end;
    end else begin
      inc(FOwner.FNextAddr,3); TempBefehl:='cjne'; TempParameter:=S+','+T+','+U;
      exit;
    end;
  end;

  {ajmp}
  If Left('ajmp ',S) then begin
    TempBefehl:='ajmp'; TempParameter:=Trim(Copy(S,6,MaxInt));
    inc(FOwner.FNextAddr,2); exit;
  end;

  {acall}
  If Left('acall ',S) then begin
    TempBefehl:='acall'; TempParameter:=Trim(Copy(S,7,MaxInt));
    inc(FOwner.FNextAddr,2); exit;
  end;

  {jb}
  If Left('jb ',S) then begin
    S:=Trim(Copy(S,4,MaxInt)); Parts(S,T); If IsBitAddr(S) then begin
      If RelDest(T,B) then SetCode([$20,GetBitAddr(S),B]) else begin
        inc(FOwner.FNextAddr,3); TempBefehl:='jb'; TempParameter:=S+','+T;
      end;
      exit;
    end;
  end;

  {jnb}
  If Left('jnb ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); Parts(S,T); If IsBitAddr(S) then begin
      If RelDest(T,B) then SetCode([$30,GetBitAddr(S),B]) else begin
        inc(FOwner.FNextAddr,3); TempBefehl:='jnb'; TempParameter:=S+','+T;
      end;
      exit;
    end;
  end;

  {jbc}
  If Left('jbc ',S) then begin
    S:=Trim(Copy(S,5,MaxInt)); Parts(S,T); If IsBitAddr(S) then begin
      If RelDest(T,B) then SetCode([$10,GetBitAddr(S),B]) else begin
        inc(FOwner.FNextAddr,3); TempBefehl:='jnb'; TempParameter:=S+','+T;
      end;
      exit;
    end;
  end;

  raise ECompilerError.Create('Unbekannter Befehl',FZeile);
end;

procedure TCommand.SolveReferenzes;
Var I : Integer;
    B : Boolean;
    C,C2 : Byte;
    S,T : String;
begin
  If length(AsmCode)<>0 then exit; B:=FindTarget(I);
  If TempBefehl='ljmp' then begin
    If B then begin SetCode([$02,I div 256,I mod 256],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='lcall' then begin
    If B then begin SetCode([$12,I div 256,I mod 256],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='sjmp' then begin
    If RelDest(C) then begin SetCode([$80,C],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='jc' then begin
    If RelDest(C) then begin SetCode([$40,C],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='jnc' then begin
    If RelDest(C) then begin SetCode([$50,C],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='jz' then begin
    If RelDest(C) then begin SetCode([$60,C],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='jnz' then begin
    If RelDest(C) then begin SetCode([$70,C],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='djnz' then begin
    S:=TempParameter; Parts(S,T); TempParameter:=T;
    If IsRx(S) then begin
      If RelDest(C) then begin
        SetCode([$D8+ord(S[2])-48,C],False); exit;
      end;
    end else begin
      If RelDest(C,3) then begin
        VarTest(S,I); SetCode([$D5,I,C],False); exit;
      end;
    end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='mov' then begin
    S:=TempParameter; Parts(S,T); TempParameter:=T;
    If S='dptr' then begin
      If FindTarget(I) then begin
        SetCode([$90,I div 256,I mod 256],False); exit;
      end else raise ECompilerError.Create('Ungltige Adreangabe',FZeile);
    end;
  end;
  If TempBefehl='cjne' then begin
    S:=TempParameter; Parts(S,T); Parts(T,TempParameter);
    B:=VarTest(T,I,'#');
    If RelDest(C,3) then begin
      {cjne a,dadr,rel} If (S='a') and VarTest(T,I) then begin SetCode([$B5,I,C],False); exit; end;
      {cjne a,#const8,rel} If (S='a') and B then begin SetCode([$B4,I,C],False); exit; end;
      {cjne Rx,#const8,rel} If IsRx(S) and B then begin SetCode([$B8+ord(S[2])-48,I,C],False); exit; end;
      {cjne @r0,#const8,rel} If (S='@r0') and B then begin SetCode([$B6,I,C],False); exit; end;
      {cjne @r1,#const8,rel} If (S='@r1') and B then begin SetCode([$B7,I,C],False); exit; end;
    end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='ajmp' then begin
    If RelDest2(C,C2) then begin SetCode([C+1,C2],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='acall' then begin
    If RelDest2(C,C2) then begin SetCode([C+17,C2],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='jb' then begin
    S:=TempParameter; Parts(S,T); TempParameter:=T;
    If RelDest(C,3) then begin SetCode([$20,GetBitAddr(S),C],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='jnb' then begin
    S:=TempParameter; Parts(S,T); TempParameter:=T;
    If RelDest(C,3) then begin SetCode([$30,GetBitAddr(S),C],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
  If TempBefehl='jbc' then begin
    S:=TempParameter; Parts(S,T); TempParameter:=T;
    If RelDest(C,3) then begin SetCode([$10,GetBitAddr(S),C],False); exit; end;
    raise ECompilerError.Create('Ungltige Sprungzielangabe',FZeile);
  end;
end;

procedure TCommand.Parts(var S, T: String);
begin
  If Pos(',',S)=0 then raise ECompilerError.Create('Ungltige Parameter',FZeile);
  T:=Trim(Copy(S,Pos(',',S)+1,MaxInt));
  S:=Trim(Copy(S,1,Pos(',',S)-1));
  If (S='') or (T='') then raise ECompilerError.Create('Ungltige Parameter',FZeile);
end;

Function TCommand.FindTarget(Var I : Integer) : Boolean;
Var X : Integer;
    S : String;
begin
  S:=TempParameter; If (S<>'') and (S[1]='#') then S:=Trim(Copy(S,2,MaxInt));
  result:=False;
  For X:=0 to FOwner.FTargetTab.Count-1 do If S=FOwner.FTargetTab[X] then begin
    I:=Integer(FOwner.FTargetTab.Objects[X]); result:=True; exit;
  end;
end;

Function TCommand.RelDest(S : String; var B : Byte; len : Byte) : Boolean;
Var I : Integer;
    Sh : ShortInt;
begin
  result:=False; If S='' then raise ECompilerError.Create('Ungltiges Sprungziel',FZeile);
  If (S[1]='+') or (S[1]='-') then begin
    If IsByteZahl(Copy(S,2,MaxInt)) then begin
      I:=StrToInt(Copy(S,2,MaxInt));
      If ((S[1]='+') and (I>127)) or ((S[1]='-') and (I>128)) then
         raise ECompilerError.Create('Das Sprungziel ist fr einen relativen Sprung zu weit entfernt',FZeile);
      If S[1]='+' then Sh:=I else Sh:=-I; B:=Byte(Sh); result:=True;
    end;
    exit;
  end;
  If IsWordZahl(S) then begin
    I:=StrToInt(S)-FAddr-len;
    If (I<-128) or (I>127) then
      raise ECompilerError.Create('Das Sprungziel ist fr einen relativen Sprung zu weit entfernt',FZeile);
    Sh:=I; B:=Byte(Sh); result:=True;
  end;
end;

Function TCommand.RelDest(var B : Byte; len : Byte) : Boolean;
Var I : Integer;
    Sh : ShortInt;
begin
  result:=False;
  If not FindTarget(I) then exit;
  I:=I-FAddr-len;
  If (I<-128) or (I>127) then
    raise ECompilerError.Create('Das Sprungziel ist fr einen relativen Sprung zu weit entfernt',FZeile);
  Sh:=I; B:=Byte(Sh); result:=True;
end;

Function TCommand.RelDest2(var B,C : Byte; len : Byte) : Boolean;
Var I : Integer;
begin
  result:=False; If not FindTarget(I) then exit;
  If (I shr 11)<>((FAddr+len) shr 11) then
    raise ECompilerError.Create('Das Sprungziel ist fr einen relativen Sprung zu weit entfernt',FZeile);
  B:=(I div 256)*32; C:=I mod 256; result:=True;
end;

Procedure TCommand.AddWords(S : String; Nr : Integer);
Var T : String;
    W : Word;
begin
  FCode:='.word '+S; FAddr:=FOwner.FNextAddr; FZeile:=Nr;
  while S<>'' do begin
    If Pos(',',S)<>0 then begin
      T:=Trim(Copy(S,1,Pos(',',S)-1)); S:=Trim(Copy(S,Pos(',',S)+1,MaxInt));
    end else begin
      T:=S; S:='';
    end;
    If IsWordZahl(T) then begin
      W:=StrToInt(T);
      SetLength(AsmCode,length(AsmCode)+2);
      AsmCode[length(AsmCode)-2]:=W div 256;
      AsmCode[length(AsmCode)-1]:=W mod 256;
      inc(FOwner.FNextAddr,2);
    end else raise ECompilerError.Create('Ungltige Konstante',FZeile);
  end;
end;

Procedure TCommand.AddBytes(S : String; Nr : Integer);
Var T : String;
begin
  FCode:='.byte '+S; FAddr:=FOwner.FNextAddr; FZeile:=Nr;
  while S<>'' do begin
    If Pos(',',S)<>0 then begin
      T:=Trim(Copy(S,1,Pos(',',S)-1)); S:=Trim(Copy(S,Pos(',',S)+1,MaxInt));
    end else begin
      T:=S; S:='';
    end;
    If IsByteZahl(T) then begin
      SetLength(AsmCode,length(AsmCode)+1);
      AsmCode[length(AsmCode)-2]:=StrToInt(T);
      inc(FOwner.FNextAddr,2);
    end else raise ECompilerError.Create('Ungltige Konstante',FZeile);
  end;
end;

{ TProgramm }

constructor TProgramm.Create;
begin
  inherited Create;
  FVarTab:=TStringlist.Create;
  FTargetTab:=TStringlist.Create;
  FRAMAddr:=TList.Create;
  FRAMValue:=TList.Create;
  FProtAddr:=TList.Create;
  SetLength(FCommand,0);
end;

destructor TProgramm.Destroy;
begin
  FreeCommands;
  FVarTab.Free; FTargetTab.Free;
  FRAMAddr.Free; FRAMValue.Free;
  FProtAddr.Free;
  inherited Destroy;
end;

procedure TProgramm.Compile(const Lines: TStrings; MemoryProtect : Boolean);
Var X : Integer;
    S : String;
begin
  FreeCommands;
  FVarTab.Clear; FTargetTab.Clear;
  FStartAddr:=0; FNextAddr:=0; FRAMFile:=''; FClearRAM:=False; FTakte:=-1;
  FRAMAddr.Clear; FRAMValue.Clear; FProtAddr.Clear; FProtectDefault:=MemoryProtect;
  For X:=0 to Lines.Count-1 do begin
    S:=Trim(Lines[X]);
    If S='' then continue;
    If (Pos(';',S)<>0) and (S[1]<>';') then S:=Copy(S,1,Pos(';',S)-1);
    S:=Trim(ExtLowerCase(S));
    If S='' then continue;
    If not ScanLine(X+1,S) then break;
  end;
  For X:=0 to Lines.Count-1 do begin
    S:=Lines[X];
    If Pos(';',S)<>0 then S:=Copy(S,1,Pos(';',S)-1);
    S:=Trim(ExtLowerCase(S));
    If S='' then continue;
    If not CompileLine(X+1,S) then break;
  end;
  For X:=0 to length(FCommand)-1 do FCommand[X].SolveReferenzes;
  If FProtectDefault then ProtectDefault;
end;

function TProgramm.ScanLine(Nr : Integer; S: String): Boolean;
begin
  result:=(S<>'.end'); If not result then exit;
  If (S[1]=';') then begin
    If Copy(Trim(Copy(S,2,MaxInt)),1,1)<>'$' then exit;
    S:=Trim(Copy(Trim(Copy(S,2,MaxInt)),2,MaxInt));
    If Left('loadram ',S) then begin FRAMFile:=Trim(Copy(S,9,MaxInt)); exit; end;
    If S='clearram' then begin FClearRAM:=True; exit; end;
    If Left('loadvar ',S) then begin
      If not LoadRamVar(Trim(Copy(S,9,MaxInt)),Nr,false) then
        raise ECompilerError.Create('Ungltige Speicherbelegung',Nr) else exit;
    end;
    If Left('protect ',S) then begin
      S:=Trim(Copy(S,9,MaxInt));
      If S='default' then begin FProtectDefault:=True; exit; end;
      If S='nodefault' then begin FProtectDefault:=False; exit; end;
      If not LoadRamVar(Trim(Copy(S,10,MaxInt)),Nr,true) then
        raise ECompilerError.Create('Ungltige Speicherbelegung',Nr) else exit;
    end;
    If Left('speedtest ',S) then begin
      S:=Copy(S,11,MaxInt); FProtectDefault:=False;
      try FTakte:=StrToInt(S); except raise ECompilerError.Create('Ungltige Angabe von Takten',Nr); end;
      exit;
    end;
    ECompilerError.Create('Unbekannte Emulator-Direktive',Nr);
  end;

  If Left('.equ ',S) then begin SetVariable(Nr,Copy(S,6,MaxInt)); exit; end;
end;

function TProgramm.CompileLine(Nr : Integer; S: String): Boolean;
begin
  result:=(S<>'.end'); If not result then exit;
  If Left('.equ ',S) then exit;

  If Left('.org ',S) then begin
    try
      FNextAddr:=StrToInt(Copy(S,6,MaxInt));
      If FStartAddr=0 then FStartAddr:=FNextAddr;
    except
      raise ECompilerError.Create('Ungltige Adresse',Nr);
    end;
    exit;
  end;

  If Pos(':',S)<>0 then begin
    FTargetTab.AddObject(Trim(Copy(S,1,Pos(':',S)-1)),TObject(FNextAddr));
    S:=Trim(Copy(S,Pos(':',S)+1,MaxInt));
    If S='' then exit;
  end;

  If Left('.word ',S) then begin
    S:=Trim(Copy(S,7,MaxInt));
    If S<>'' then begin
      SetLength(FCommand,length(FCommand)+1);
      FCommand[length(FCommand)-1]:=TCommand.Create(self);
      FCommand[length(FCommand)-1].AddWords(S,Nr);
    end;
    exit;
  end;
  If Left('.byte ',S) then begin
    S:=Trim(Copy(S,7,MaxInt));
    If S<>'' then begin
      SetLength(FCommand,length(FCommand)+1);
      FCommand[length(FCommand)-1]:=TCommand.Create(self);
      FCommand[length(FCommand)-1].AddBytes(S,Nr);
    end;
    exit;
  end;

  SetLength(FCommand,length(FCommand)+1);
  FCommand[length(FCommand)-1]:=TCommand.Create(self);
  FCommand[length(FCommand)-1].Compile(S,Nr);
end;

procedure TProgramm.SetVariable(Nr: Integer; S: String);
Var Name, Value : String;
    I : Integer;
begin
  If Pos(',',S)=0 then raise ECompilerError.Create('Ungltige Variablendeklaration',Nr);
  Name:=Trim(Copy(S,1,Pos(',',S)-1)); Value:=Trim(Copy(S,Pos(',',S)+1,MaxInt));
  If Pos(' ',Name)<>0 then raise ECompilerError.Create('Ungltiger Variablenname',Nr);
  try I:=StrToInt(Value) except raise ECompilerError.Create('Ungltiger Zahlenwert',Nr); end;
  If (I<0) or (I>255) then raise ECompilerError.Create('Zahlenwert auerhalb des gltigen Bereiches',Nr);
  FVarTab.AddObject(Name,TObject(I));
end;

procedure TProgramm.FreeCommands;
Var X : Integer;
begin
  For X:=0 to length(FCommand)-1 do FCommand[X].Free;
  SetLength(FCommand,0);
end;

function TProgramm.GetListing: TStrings;
Var X,Y,M,C,D : Integer;
    S : String;
begin
  result:=TStringlist.Create;
  result.Add(PrgName+' Version '+Version);
  result.Add('');
  If FVarTab.Count>0 then begin
    result.Add('VARIABLEN-TABELLE:');
    M:=0;
    For X:=0 to FVarTab.Count-1 do If length(FVarTab[X])>M then M:=length(FVarTab[X]);
    For X:=0 to FVarTab.Count-1 do begin
      S:=FVarTab[X]; while length(S)<M do S:=S+' ';
      result.Add(S+' = 0x'+IntToHex(Integer(FVarTab.Objects[X]),2)+' ('+IntToStr(Integer(FVarTab.Objects[X]))+')');
    end;
    result.Add('');
  end;
  If FTargetTab.Count>0 then begin
    result.Add('SPRUNGMARKEN-TABELLE:');
    M:=0;
    For X:=0 to FTargetTab.Count-1 do If length(FTargetTab[X])>M then M:=length(FTargetTab[X]);
    For X:=0 to FTargetTab.Count-1 do begin
      S:=FTargetTab[X]; while length(S)<M do S:=S+' ';
      result.Add(S+' : 0x'+IntToHex(Integer(FTargetTab.Objects[X]),2)+' ('+IntToStr(Integer(FTargetTab.Objects[X]))+')');
    end;
    result.Add('');
  end;
  M:=0; C:=result.Count;
  For X:=0 to length(FCommand)-1 do begin
    S:=IntToHex(FCommand[X].Addr,4)+': ';
    For Y:=0 to length(FCommand[X].AsmCode)-1 do S:=S+IntToHex(FCommand[X].AsmCode[Y],2)+' ';
    If length(S)>M then M:=length(S);
    result.Add(S);
  end;
  D:=0;
  For X:=0 to length(FCommand)-1 do begin
    S:=result[X+C+D];
    while length(S)<M do S:=S+' ';
    For Y:=0 to FTargetTab.Count-1 do If Integer(FTargetTab.Objects[Y])=FCommand[x].Addr then begin
      result.Insert(X+C+D,FTargetTab[Y]+':'); inc(D);
    end;
    result[X+C+D]:=S+FCommand[X].Code;
  end;
end;

function TProgramm.GetAsmCode(MiniOS : Boolean): TStream;
Var B,X : Integer;
begin
  result:=TMemoryStream.Create; B:=0;
  If length(FCommand)=0 then exit;
  For X:=0 to length(FCommand)-1 do If FCommand[X].Addr+length(FCommand[X].AsmCode)>B then B:=FCommand[X].Addr+length(FCommand[X].AsmCode);

  If MiniOS then begin If B<3+15*8 then B:=3+15*8; end;

  TMemoryStream(result).SetSize(B);
  FillChar(TMemoryStream(result).Memory^,B,0);

  If MiniOS then begin
    For X:=0 to 15 do begin
      result.Position:=3+X*8;
      Case X of
        6,7,12,14 : ;
        else B:=$32; result.Write(B,1);
      end;
    end;
  end;

  For X:=0 to length(FCommand)-1 do begin
    result.Position:=FCommand[X].Addr;
    result.Write(FCommand[X].AsmCode[0],length(FCommand[X].AsmCode));
  end;
  result.Position:=0;
end;

function TProgramm.GetAsmCode2: TStrings;
Var X,Y,CodeAddr : Integer;
    S : String;
begin
  result:=TStringlist.Create;
  If length(FCommand)=0 then exit;

  S:='';
  For Y:=0 to length(FCommand[0].AsmCode)-1 do S:=S+IntToHex(FCommand[0].AsmCode[Y],2);
  CodeAddr:=FCommand[0].Addr+length(FCommand[0].AsmCode);

  For X:=1 to length(FCommand)-1 do begin
    while CodeAddr<FCommand[X].Addr do begin
      S:=S+'00'; inc(CodeAddr);
      If length(S)>240 then begin result.Add(S); S:=''; end;
    end;
    For Y:=0 to length(FCommand[X].AsmCode)-1 do S:=S+IntToHex(FCommand[X].AsmCode[Y],2);
    If length(S)>240 then begin result.Add(S); S:=''; end;
    inc(CodeAddr,length(FCommand[X].AsmCode));
  end;

  If length(S)>0 then result.Add(S);
end;

function TProgramm.GetCommandCount: Integer;
begin
  result:=length(FCommand);
end;

function TProgramm.GetCommand(I: Integer): TCommand;
begin
  If (I<0) or (I>=length(FCommand)) then result:=nil else result:=FCommand[I];
end;

procedure TProgramm.Parts(var S, T: String; Nr : Integer);
begin
  If Pos(',',S)=0 then raise ECompilerError.Create('Ungltige Parameter',Nr);
  T:=Trim(Copy(S,Pos(',',S)+1,MaxInt));
  S:=Trim(Copy(S,1,Pos(',',S)-1));
  If (S='') or (T='') then raise ECompilerError.Create('Ungltige Parameter',Nr);
end;

function TProgramm.LoadRamVar(S: String; Nr : Integer; Prot : Boolean): Boolean;
Var T : String;
    I,J : Integer;
begin
  result:=False;
  If Prot then begin
    try I:=StrToInt(S); except exit; end;
    If (I<0) or (I>255) then exit;
    If FProtAddr.IndexOf(Pointer(I))=-1 then FProtAddr.Add(Pointer(I));
  end else begin
    Parts(S,T,Nr);
    try I:=StrToInt(S); J:=StrToInt(T); except exit; end;
    If (I<0) or (I>255) or (J<0) or (J>255) then exit;
    If FRAMAddr.IndexOf(Pointer(I))=-1 then begin
      FRAMAddr.Add(Pointer(I));
      FRAMValue.Add(Pointer(J));
    end;  
  end;
  result:=True;
end;

procedure TProgramm.ProtectDefault;
Var X : Integer;
begin
  For X:=$18 to $1F do If FProtAddr.IndexOf(Pointer(X))=-1 then FProtAddr.Add(Pointer(X));
  For X:=$20 to $21 do If FProtAddr.IndexOf(Pointer(X))=-1 then FProtAddr.Add(Pointer(X));
  For X:=$6C to $79 do If FProtAddr.IndexOf(Pointer(X))=-1 then FProtAddr.Add(Pointer(X));
  For X:=$59 to $59 do If FProtAddr.IndexOf(Pointer(X))=-1 then FProtAddr.Add(Pointer(X));
  For X:=$90 to $92 do If FProtAddr.IndexOf(Pointer(X))=-1 then FProtAddr.Add(Pointer(X));
end;

end.
