unit ExtDlg;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, ComCtrls, ToolWin, StdCtrls, ExtCtrls;

type
  TExtRAMForm = class(TForm)
    ToolBar: TToolBar;
    ToolButton1: TToolButton;
    Tab: TStringGrid;
    Wahl: TComboBox;
    Label1: TLabel;
    Timer: TTimer;
    ToolButton3: TToolButton;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure ToolButton1Click(Sender: TObject);
    procedure WahlChange(Sender: TObject);
    procedure TabDblClick(Sender: TObject);
    procedure TabKeyPress(Sender: TObject; var Key: Char);
    procedure TabMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure TimerTimer(Sender: TObject);
    procedure ToolButton3Click(Sender: TObject);
  private
    { Private-Deklarationen }
    HintWnd : THintWindow;
    LastX,LastY : Integer;
    Data : Boolean;
    Function GetDisCode(X,Y : Integer) : String;
  public
    { Public-Deklarationen }
    Procedure ReadData;
  end;

var
  ExtRAMForm: TExtRAMForm = nil;

implementation

uses MainUnit, InputDlg, Tools, Decompiler;

{$R *.DFM}

procedure TExtRAMForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  ExtRAMForm:=nil;
  Action:=caFree;
  HintWnd.Free;
end;

procedure TExtRAMForm.FormCreate(Sender: TObject);
Var X : Integer;
begin
  Data:=True;
  For X:=0 to 255 do
    Wahl.Items.Add('0x'+IntToHex(X*256,4)+'-0x'+IntToHex((X+1)*256-1,4));
  Wahl.ItemIndex:=1;
  Tab.Cells[0,0]:='Adresse';
  For X:=0 to 15 do Tab.Cells[X+1,0]:='+'+IntToHex(X,1);
  ReadData;
  If MainForm.WindowStyle.ItemIndex>=0 then
    BoundsRect:=MainForm.SetupData.WindowSetup[MainForm.WindowStyle.ItemIndex].ExtRAMWnd;
  HintWnd:=THintWindow.Create(self);
  LastX:=-1; LastY:=-1;
end;

procedure TExtRAMForm.ReadData;
Var X,Y,Z : Integer;
begin
  ToolButton1.Down:=MainForm.System8051.ALU.Reg.XMAP;
  For X:=0 to 16 do Tab.Cells[0,X+1]:='0x'+IntToHex(Wahl.ItemIndex*256+X*16,4);
  For Y:=0 to 15 do for X:=0 to 15 do begin
    Z:=Wahl.ItemIndex*256+Y*16+X;
    If Data then begin
      If (Z>=$FF00) and ToolButton1.Down then
        Tab.Cells[X+1,Y+1]:=IntToHex(MainForm.System8051.XRAM[Z-$FF00],2)
      else
        Tab.Cells[X+1,Y+1]:=IntToHex(MainForm.System8051.Ext[Z],2);
    end else Tab.Cells[X+1,Y+1]:=IntToHex(MainForm.System8051.Code[Z],2);
  end;
end;

procedure TExtRAMForm.FormResize(Sender: TObject);
Var X,Y : Integer;
begin
  Tab.ColWidths[0]:=50;
  Y:=(Tab.ClientWidth-65) div 16; If Y<25 then Y:=25;
  For X:=1 to 16 do Tab.ColWidths[X]:=Y;
end;

procedure TExtRAMForm.ToolButton1Click(Sender: TObject);
begin
  MainForm.System8051.ALU.Reg.XMAP:=ToolButton1.Down;
  MainForm.ReadData;
end;

procedure TExtRAMForm.WahlChange(Sender: TObject);
begin
  ReadData;
end;

procedure TExtRAMForm.TabDblClick(Sender: TObject);
Var Z,I : Integer;
begin
  If (Tab.Row<1) or (Tab.Col<1) then exit;
  Z:=Wahl.ItemIndex*256+(Tab.Row-1)*16+(Tab.Col-1);
  If Data then begin
    If (Z>=$FF00) and ToolButton1.Down then
      I:=MainForm.System8051.XRAM[Z-$FF00]
    else
      I:=MainForm.System8051.Ext[Z];
  end else I:=MainForm.System8051.Code[Z];
  If InputNumber(I,1) then begin
    If Data then begin
      If (Z>=$FF00) and ToolButton1.Down then
        MainForm.System8051.XRAM[Z-$FF00]:=I
      else begin
        MainForm.System8051.Ext[Z]:=I;
        If ((Z<=$7F00) and MainForm.System8051.ALU.SingleMemory1) or ((Z>$7F00) and MainForm.System8051.ALU.SingleMemory2) then
          MainForm.System8051.Code[Z]:=I;
      end;
    end else begin
      MainForm.System8051.Code[Z]:=I;
      If ((Z<=$7F00) and MainForm.System8051.ALU.SingleMemory1) or ((Z>$7F00) and MainForm.System8051.ALU.SingleMemory2) then
        MainForm.System8051.Ext[Z]:=I;
    end;
    MainForm.ReadData;
  end;
end;

procedure TExtRAMForm.TabKeyPress(Sender: TObject; var Key: Char);
Var B : Integer;
    C : Char;
begin
  C:=ExtUpperCase(Key)[1]; B:=-1;
  If (C>='0') and (C<='9') then B:=ord(C)-48;
  If (C>='A') and (C<='F') then B:=ord(C)-65+10;
  If B>=0 then Tab.TopRow:=1+B;
end;

procedure TExtRAMForm.TabMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
Var P : TPoint;
    R : TRect;
    S : String;
begin
  P:=Point(X,Y);
  Tab.MouseToCell(P.X,P.Y,X,Y);
  If (X>0) and (Y>0) and ((LastX<>X) or (LastY<>Y)) then begin
    P:=ClientToScreen(P); LastX:=X; LastY:=Y;
    S:=GetDisCode(X,Y);
    R:=HintWnd.CalcHintRect(200,S,nil);
    inc(R.Left,P.X); inc(R.Right,P.X); inc(R.Top,P.Y); inc(R.Bottom,P.Y);
    HintWnd.ActivateHint(R,S);
  end;
end;

procedure TExtRAMForm.TimerTimer(Sender: TObject);
Var P : TPoint;
    X,Y : Integer;
begin
  If HintWnd.Visible then begin
    GetCursorPos(P); P:=ScreenToClient(P); Tab.MouseToCell(P.X,P.Y,X,Y);
    If (X<=0) or (Y<=0) then begin
      HintWnd.ReleaseHandle; LastX:=-1; LastY:=-1;
    end;
  end;
end;

function TExtRAMForm.GetDisCode(X, Y: Integer): String;
Var St : TMemoryStream;
    Z,I : Integer;
    B : Byte;
    S : String;
begin
  result:=''; St:=TMemoryStream.Create; dec(X); dec(Y);
  try
    For I:=1 To 3 do begin
      Z:=Wahl.ItemIndex*256+Y*16+X;
      If Data then begin
        If (Z>=$FF00) and ToolButton1.Down then
          B:=MainForm.System8051.XRAM[Z-$FF00]
        else
          B:=MainForm.System8051.Ext[Z];
      end else B:=MainForm.System8051.Code[Z];
      St.Write(B,1);
      inc(X); If X=15 then begin inc(Y); X:=0; end;
      If Y=16 then break;
    end;
    St.Position:=0; Z:=DecompileOp(St,S); St.Position:=0;
    For I:=1 to Z do begin St.Read(B,1); result:=result+IntToHex(B,2)+' '; end;
    result:=result+'= '+S;
  finally
    St.Free;
  end;
end;

procedure TExtRAMForm.ToolButton3Click(Sender: TObject);
begin
  If Data
    then ToolButton3.Caption:='Datenspeicher'
    else ToolButton3.Caption:='Codespeicher';
  ToolButton3.Hint:='Umschalten zum '+ToolButton3.Caption;
  Data:=not Data;
  ToolButton1.Visible:=Data;
  ReadData;
end;

end.
