unit RAMDlg;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, ImgList, ComCtrls, ToolWin, StdCtrls;

type
  TRAMForm = class(TForm)
    Tab: TStringGrid;
    ToolBar: TToolBar;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ImageList: TImageList;
    ToolButton3: TToolButton;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    HighRAM: TComboBox;
    ToolButton4: TToolButton;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure TabDblClick(Sender: TObject);
    procedure TabKeyPress(Sender: TObject; var Key: Char);
    procedure ButtonWork(Sender: TObject);
    procedure HighRAMChange(Sender: TObject);
  private
    { Private-Deklarationen }
  public
    { Public-Deklarationen }
    Procedure ReadData;
  end;

var
  RAMForm: TRAMForm = nil;

implementation

uses MainUnit, InputDlg, Tools, PrgConst;

{$R *.DFM}

procedure TRAMForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  RAMForm:=nil;
  Action:=caFree;
end;

procedure TRAMForm.FormCreate(Sender: TObject);
Var X : Integer;
begin
  HighRAM.ItemIndex:=0;
  Tab.Cells[0,0]:='Adresse';
  Tab.Cells[1,0]:='Wert';
  Tab.RowCount:=257;
  For X:=0 to 255 do Tab.Cells[0,X+1]:='0x'+IntToHex(X,2);
  ReadData;
  If MainForm.WindowStyle.ItemIndex>=0 then
    BoundsRect:=MainForm.SetupData.WindowSetup[MainForm.WindowStyle.ItemIndex].RAMWnd;
end;

procedure TRAMForm.ReadData;
Var X : Integer;
    S : String;
begin
  If HighRAM.ItemIndex=0 then begin
    For X:=0 to 255 do Tab.Cells[1,X+1]:=
        '0x'+IntToHex(MainForm.System8051.Data[X],2)+' ('+IntToStr(MainForm.System8051.Data[X])+')';
  end else begin
    For X:=0 to 127 do Tab.Cells[1,X+1]:=
        '0x'+IntToHex(MainForm.System8051.Data[X],2)+' ('+IntToStr(MainForm.System8051.Data[X])+')';
    For X:=128 to 255 do begin
      Tab.Cells[1,X+1]:=
        '0x'+IntToHex(MainForm.System8051.SFRArea[X],2)+' ('+IntToStr(MainForm.System8051.SFRArea[X])+')';
      S:='';
      Case X of
        RegA      : S:='ACC';
        RegB      : S:='B';
        RegPSW    : S:='PSW';
        RegSP     : S:='SP';
        RegDPL    : S:='DPL';
        RegDPH    : S:='DPH';
        RegDPSEL  : S:='DPSEL';
        RegSYSCON : S:='SYSCON';
        RegP0     : S:='P0';
        RegP1     : S:='P1';
        RegP2     : S:='P2';
        RegP3     : S:='P3';
        RegP4     : S:='P4';
        RegP5     : S:='P5';
        RegIEN0   : S:='IEN0';
        RegIEN1   : S:='IEN1';
        RegIP0    : S:='IP0';
        RegIP1    : S:='IP1';
        RegPCON   : S:='PCON';
        RegSBUF   : S:='SBUF';
        RegSCON   : S:='SCON';
        RegTCON   : S:='TCON';
        RegTMOD   : S:='TMOD';
        RegT2CON  : S:='T2CON';
        RegT2MOD  : S:='T2MOD';
        RegADCON0 : S:='ADCON0';
        RegADCON1 : S:='ADCON1';
        RegADDATH : S:='ADDATH';
        RegADDATL : S:='ADDATL';
        RegITCON  : S:='ITCON';
        RegTRCON  : S:='TRCON';
        RegBCON   : S:='BCON';
        RegCT2CON : S:='CT2CON';
        RegCCIR   : S:='CCIR';
        RegTL0    : S:='TL0';
        RegTH0    : S:='TH0';
        RegTL1    : S:='TL1';
        RegTH1    : S:='TH1';
        RegCT1CON : S:='CT1CON';
        RegCCPL   : S:='CCPL';
        RegCCPH   : S:='CCPH';
        RegCCL0   : S:='CCL0';
        RegCCH0   : S:='CCH0';
        RegCCL1   : S:='CCL1';
        RegCCH1   : S:='CCH1';
        RegCCL2   : S:='CCL2';
        RegCCH2   : S:='CCH2';
        RegCT1OFL : S:='CT1OFL';
        RegCT1OFH : S:='CT1OFH';
        RegCMSEL0 : S:='CMSEL0';
        RegCMSEL1 : S:='CMSEL1';
      end;
      If S<>'' then Tab.Cells[1,X+1]:=Tab.Cells[1,X+1]+' ='+S;
    end;
  end;
  ToolButton4.Down:=MainForm.System8051.ALU.Reg.RMAP;
end;

procedure TRAMForm.FormResize(Sender: TObject);
begin
  Tab.ColWidths[0]:=Tab.ClientWidth div 3;
  Tab.ColWidths[1]:=2*Tab.ClientWidth div 3;
end;

procedure TRAMForm.TabDblClick(Sender: TObject);
Var I : Integer;
begin
  If Tab.Row<1 then exit;
  If (HighRAM.ItemIndex=0) or (Tab.Row-1<128) then
    I:=MainForm.System8051.Data[Tab.Row-1]
  else
    I:=MainForm.System8051.SFRArea[Tab.Row-1];
  If InputNumber(I,1) then begin
    If (HighRAM.ItemIndex=0) or (Tab.Row-1<128) then
      MainForm.System8051.Data[Tab.Row-1]:=I
    else
      MainForm.System8051.SFRArea[Tab.Row-1]:=I;
    MainForm.ReadData;
  end;
end;

procedure TRAMForm.TabKeyPress(Sender: TObject; var Key: Char);
Var B : Integer;
    C : Char;
begin
  C:=ExtUpperCase(Key)[1]; B:=-1;
  If (C>='0') and (C<='9') then B:=ord(C)-48;
  If (C>='A') and (C<='F') then B:=ord(C)-65+10;
  If B>=0 then Tab.TopRow:=1+(B*16);
end;

procedure TRAMForm.ButtonWork(Sender: TObject);
Var FSt : TFileStream;
begin
  Case (Sender as TToolButton).Tag of
    0 : If MessageDlg('Soll wirklich der gesamte Variablenspeicher gelscht werden ?',mtConfirmation,[mbYes,mbNo],0)=mrYes then begin
          MainForm.System8051.Data.Clear;
        end;
    1 : If OpenDialog.Execute then begin
          FSt:=TFileStream.Create(OpenDialog.FileName,fmOpenRead);
          MainForm.System8051.Data.Clear;
          MainForm.System8051.Data.LoadData(FSt,0);
          FSt.Free;
        end;
    2 : If SaveDialog.Execute then begin
          FSt:=TFileStream.Create(SaveDialog.FileName,fmCreate);
          MainForm.System8051.Data.SaveData(FSt,0);
          FSt.Free;
        end;
    3 : MainForm.System8051.ALU.Reg.RMAP:=not MainForm.System8051.ALU.Reg.RMAP;
  end;
  MainForm.ReadData;
end;

procedure TRAMForm.HighRAMChange(Sender: TObject);
begin
  MainForm.ReadData;
end;

end.
